#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import re

from .style_trait import StyleTrait


HEX_COLOR = re.compile('#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{8})$')


class Color(tuple):
    """ A tuple subclass which represents an rgba color value.

    """
    __slots__ = ()

    @classmethod
    def from_string(cls, color_string):
        """ Convert a string to an enaml Color instance.

        Arguments
        ---------
        color_string : str
            String describing color in hexadecimal format or a CSS color
            name. The hex formats supported are #rrggbb and #rrggbbaa.

        Returns
        -------
        color : Instance(Color)
            The rgba color tuple representation in enaml.

        Notes
        -----
        The color names supported are the 147 specified by the SVG
        color standard: http://www.w3.org/TR/SVG/types.html#ColorKeywords
        Plus an Enaml defined 'error' color. All of the named colors are 
        fully opaque.

        """
        color_string = color_string.strip()
        match = HEX_COLOR.match(color_string)
        if match:
            hex_str = match.group(1)
            r = int(hex_str[:2], 16)
            g = int(hex_str[2:4], 16)
            b = int(hex_str[4:6], 16)
            if len(hex_str) == 8:
                a = int(hex_str[6:8], 16)
            else:
                a = 255
            res = cls(r, g, b, a)
        else:
            res = cls.color_map.get(color_string)
            if res is None:
                res = cls(-1, -1, -1, -1)
        return res
    
    @classmethod
    def from_frgba(cls, r, g, b, a):
        r = int(255 * r)
        g = int(255 * g)
        b = int(255 * b)
        a = int(255 * a)
        return cls(r, g, b, a)
    
    @classmethod
    def from_frgb(cls, r, g, b):
        return cls.from_frgba(r, g, b, 1.0)

    def __new__(cls, r=-1, g=-1, b=-1, a=255):
        vals = (r, g, b, a)
        for val in vals:
            if not isinstance(val, int) or not (-1 <= val <= 255):
                raise ValueError('rgba values must be -1 <= val <= 255')
        return tuple.__new__(cls, vals)
    
    def __nonzero__(self):
        return all(val >= 0 for val in self)
        
    def __repr__(self):
        if not self:
            inner = '<invalid>'
        else:
            inner = 'r=%s, g=%s, b=%s, a=%s' % self
        return 'Color(%s)' % inner

    def __str__(self):
        return self.__repr__()

    def clone(self, r=None, g=None, b=None, a=None):
        these = (r, g, b, a)
        vals = ((this if this is not None else other) 
                for (this, other) in zip(these, self))
        return Color(*vals)

    @property
    def r(self):
        return self[0]
    
    @property
    def g(self):
        return self[1]
    
    @property
    def b(self):
        return self[2]
    
    @property
    def a(self):
        return self[3]


#: All 147 CSS colors plus some of our own.
Color.color_map = {
    'aliceblue': Color(240, 248, 255),
    'antiquewhite': Color(250, 235, 215),
    'aqua': Color(0, 255, 255),
    'aquamarine': Color(127, 255, 212),
    'azure': Color(240, 255, 255),
    'beige': Color(245, 245, 220),
    'bisque': Color(255, 228, 196),
    'black': Color(0, 0, 0),
    'blanchedalmond': Color(255, 235, 205),
    'blue': Color(0, 0, 255),
    'blueviolet': Color(138, 43, 226),
    'brown': Color(165, 42, 42),
    'burlywood': Color(222, 184, 135),
    'cadetblue': Color(95, 158, 160),
    'chartreuse': Color(127, 255, 0),
    'chocolate': Color(210, 105, 30),
    'coral': Color(255, 127, 80),
    'cornflowerblue': Color(100, 149, 237),
    'cornsilk': Color(255, 248, 220),
    'crimson': Color(220, 20, 60),
    'cyan': Color(0, 255, 255),
    'darkblue': Color(0, 0, 139),
    'darkcyan': Color(0, 139, 139),
    'darkgoldenrod': Color(184, 134, 11),
    'darkgray': Color(169, 169, 169),
    'darkgreen': Color(0, 100, 0),
    'darkgrey': Color(169, 169, 169),
    'darkkhaki': Color(189, 183, 107),
    'darkmagenta': Color(139, 0, 139),
    'darkolivegreen': Color(85, 107, 47),
    'darkorange': Color(255, 140, 0),
    'darkorchid': Color(153, 50, 204),
    'darkred': Color(139, 0, 0),
    'darksalmon': Color(233, 150, 122),
    'darkseagreen': Color(143, 188, 143),
    'darkslateblue': Color(72, 61, 139),
    'darkslategray': Color(47, 79, 79),
    'darkslategrey': Color(47, 79, 79),
    'darkturquoise': Color(0, 206, 209),
    'darkviolet': Color(148, 0, 211),
    'deeppink': Color(255, 20, 147),
    'deepskyblue': Color(0, 191, 255),
    'dimgray': Color(105, 105, 105),
    'dimgrey': Color(105, 105, 105),
    'dodgerblue': Color(30, 144, 255),
    'firebrick': Color(178, 34, 34),
    'floralwhite': Color(255, 250, 240),
    'forestgreen': Color(34, 139, 34),
    'fuchsia': Color(255, 0, 255),
    'gainsboro': Color(220, 220, 220),
    'ghostwhite': Color(248, 248, 255),
    'gold': Color(255, 215, 0),
    'goldenrod': Color(218, 165, 32),
    'gray': Color(128, 128, 128),
    'grey': Color(128, 128, 128),
    'green': Color(0, 128, 0),
    'greenyellow': Color(173, 255, 47),
    'honeydew': Color(240, 255, 240),
    'hotpink': Color(255, 105, 180),
    'indianred': Color(205, 92, 92),
    'indigo': Color(75, 0, 130),
    'ivory': Color(255, 255, 240),
    'khaki': Color(240, 230, 140),
    'lavender': Color(230, 230, 250),
    'lavenderblush': Color(255, 240, 245),
    'lawngreen': Color(124, 252, 0),
    'lemonchiffon': Color(255, 250, 205),
    'lightblue': Color(173, 216, 230),
    'lightcoral': Color(240, 128, 128),
    'lightcyan': Color(224, 255, 255),
    'lightgoldenrodyellow': Color(250, 250, 210),
    'lightgray': Color(211, 211, 211),
    'lightgreen': Color(144, 238, 144),
    'lightgrey': Color(211, 211, 211),
    'lightpink': Color(255, 182, 193),
    'lightsalmon': Color(255, 160, 122),
    'lightseagreen': Color(32, 178, 170),
    'lightskyblue': Color(135, 206, 250),
    'lightslategray': Color(119, 136, 153),
    'lightslategrey': Color(119, 136, 153),
    'lightsteelblue': Color(176, 196, 222),
    'lightyellow': Color(255, 255, 224),
    'lime': Color(0, 255, 0),
    'limegreen': Color(50, 205, 50),
    'linen': Color(250, 240, 230),
    'magenta': Color(255, 0, 255),
    'maroon': Color(128, 0, 0),
    'mediumaquamarine': Color(102, 205, 170),
    'mediumblue': Color(0, 0, 205),
    'mediumorchid': Color(186, 85, 211),
    'mediumpurple': Color(147, 112, 219),
    'mediumseagreen': Color(60, 179, 113),
    'mediumslateblue': Color(123, 104, 238),
    'mediumspringgreen': Color(0, 250, 154),
    'mediumturquoise': Color(72, 209, 204),
    'mediumvioletred': Color(199, 21, 133),
    'midnightblue': Color(25, 25, 112),
    'mintcream': Color(245, 255, 250),
    'mistyrose': Color(255, 228, 225),
    'moccasin': Color(255, 228, 181),
    'navajowhite': Color(255, 222, 173),
    'navy': Color(0, 0, 128),
    'none': Color(),
    'oldlace': Color(253, 245, 230),
    'olive': Color(128, 128, 0),
    'olivedrab': Color(107, 142, 35),
    'orange': Color(255, 165, 0),
    'orangered': Color(255, 69, 0),
    'orchid': Color(218, 112, 214),
    'palegoldenrod': Color(238, 232, 170),
    'palegreen': Color(152, 251, 152),
    'paleturquoise': Color(175, 238, 238),
    'palevioletred': Color(219, 112, 147),
    'papayawhip': Color(255, 239, 213),
    'peachpuff': Color(255, 218, 185),
    'peru': Color(205, 133, 63),
    'pink': Color(255, 192, 203),
    'plum': Color(221, 160, 221),
    'powderblue': Color(176, 224, 230),
    'purple': Color(128, 0, 128),
    'red': Color(255, 0, 0),
    'rosybrown': Color(188, 143, 143),
    'royalblue': Color(65, 105, 225),
    'saddlebrown': Color(139, 69, 19),
    'salmon': Color(250, 128, 114),
    'sandybrown': Color(244, 164, 96),
    'seagreen': Color(46, 139, 87),
    'seashell': Color(255, 245, 238),
    'sienna': Color(160, 82, 45),
    'silver': Color(192, 192, 192),
    'skyblue': Color(135, 206, 235),
    'slateblue': Color(106, 90, 205),
    'slategray': Color(112, 128, 144),
    'slategrey': Color(112, 128, 144),
    'snow': Color(255, 250, 250),
    'springgreen': Color(0, 255, 127),
    'steelblue': Color(70, 130, 180),
    'tan': Color(210, 180, 140),
    'teal': Color(0, 128, 128),
    'thistle': Color(216, 191, 216),
    'tomato': Color(255, 99, 71),
    'turquoise': Color(64, 224, 208),
    'violet': Color(238, 130, 238),
    'wheat': Color(245, 222, 179),
    'white': Color(255, 255, 255),
    'whitesmoke': Color(245, 245, 245),
    'yellow': Color(255, 255, 0),
    'yellowgreen': Color(154, 205, 50), 

    # Custom Enaml colors
    'error': Color(249, 167, 167),
    'unacceptable': Color(249, 249, 167),
    'none': Color(-1, -1, -1, -1),

}


class ColorTrait(StyleTrait):

    def create_default(self, obj, name):
        return Color()

    def convert(self, obj, name, value):
        if isinstance(value, basestring):
            try:
                color = Color.from_string(value)
            except (TypeError, ValueError):
                color = self.create_default(obj,  name)
        elif isinstance(value, Color):
            color = value
        elif isinstance(value, (tuple, list)):
            try:
                color = Color(*value)
            except (TypeError, ValueError):
                color = self.create_default(obj, name)
        else:
            color = self.create_default(obj, name)
        return color




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt_widget_component import QtWidgetComponent

from ...components.constraints_widget import AbstractTkConstraintsWidget


class QtConstraintsWidget(QtWidgetComponent, AbstractTkConstraintsWidget):
    """ A Qt4 implementation of ConstraintsWidget.

    """
    pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from unicodedata import category as ucategory

import wx

from .wx_control import WXControl

from ...components.field import AbstractTkField
from ...guard import guard


class WXFieldValidator(wx.PyValidator):
    """ A wx.PyValidator implementation which proxies the work to the 
    Enaml validator installed on the shell component.

    """
    def __init__(self, validator):
        """ Initialize a WXFieldValidator

        Parameters
        ----------
        validator : AbstractValidator
            An instance of the Enaml AbstractValidator.

        """
        super(WXFieldValidator, self).__init__()
        self.validator = validator
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_TEXT_ENTER, self.OnTextEnter)
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)

    def Clone(self):
        """ A required wx.PyValidator method which returns a clone of
        this object by calling the constructor with the internal
        Enaml validator instance.

        """
        return self.__class__(self.validator)

    def OnChar(self, event):
        """ An event handler which intercepts the char events before the
        control is updated and runs the new string through the validator.

        """
        # We only need to validate the string if the key pressed can
        # actually have an effect on the displayable text. The following
        # clause simply filters any control keys which generate char
        # events, such as the arrow keys.
        uchar = unichr(event.GetUnicodeKey())
        if ucategory(uchar).startswith('C'):
            event.Skip()
            return

        # We only skip the char event if the new text validates as 
        # INTERMEDIATE or ACCEPTABLE. Otherwise, we kill the event
        # so that the control is not visibly updated.
        v = self.validator
        window = self.GetWindow()
        current = window.GetValue()
        idx = window.GetInsertionPoint()
        new = current[:idx] + uchar + current[idx:]
        if v.validate(new) != v.INVALID:
            event.Skip()

    def OnTextEnter(self, event):
        """ An event handler which intercepts the enter pressed event
        and only skips it if the text is valid or can be made valid.

        """
        v = self.validator
        window = self.GetWindow()
        current = window.GetValue()
        if v.validate(current) == v.ACCEPTABLE:
            event.Skip()
        else:
            new = v.normalize(current)
            if v.validate(new) == v.ACCEPTABLE:
                window.ChangeValue(new)
                event.Skip()
    
    def OnKillFocus(self, event):
        """ An event handler which intercepts the lost focus event
        and will attempt to make the text valid before forwarding 
        the event.

        """
        v = self.validator
        window = self.GetWindow()
        current = window.GetValue()
        if v.validate(current) != v.ACCEPTABLE:
            new = v.normalize(current)
            if v.validate(new) == v.ACCEPTABLE:
                window.ChangeValue(new)
        event.Skip()


class wxLineEdit(wx.TextCtrl):
    """ A wx.TextCtrl subclass which is similar to a QLineEdit in terms
    of features and validation behavior.

    """
    def __init__(self, *args, **kwargs):
        super(wxLineEdit, self).__init__(*args, **kwargs)
        self._placeholder_text = ''
        self._placeholder_active = False
        self._user_fgcolor = None
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)
        self.Bind(wx.EVT_SET_FOCUS, self.OnSetFocus)

    #--------------------------------------------------------------------------
    # Private API
    #--------------------------------------------------------------------------
    def _UpdatePlaceholderDisplay(self):
        """ Updates the display with the placeholder text if no text
        is currently set for the control.
        
        """ 
        if not self.GetValue() and self._placeholder_text:
            self.ChangeValue(self._placeholder_text)
            color = wx.Color(95, 95, 95)
            super(wxLineEdit, self).SetForegroundColour(color)
            self._placeholder_active = True

    def _RemovePlaceholderDisplay(self):
        """ Removes the placeholder text if it is currently active.

        """
        if self._placeholder_active:
            self.ChangeValue('')
            color = self._user_fgcolor or wx.Color(0, 0, 0)
            super(wxLineEdit, self).SetForegroundColour(color)
            self._placeholder_active = False

    #--------------------------------------------------------------------------
    # Event Handlers
    #--------------------------------------------------------------------------
    def OnKillFocus(self, event):
        """ Refreshes the placeholder display when the control loses
        focus.

        """
        self._UpdatePlaceholderDisplay()
        event.Skip()

    def OnSetFocus(self, event):
        """ Removes the placeholder display when the control receives
        focus.

        """
        self._RemovePlaceholderDisplay()
        event.Skip()
    
    #--------------------------------------------------------------------------
    # Public API
    #--------------------------------------------------------------------------
    def SetPlaceHolderText(self, placeholder_text):
        """ Sets the placeholder text to the given value. Pass an empty
        string to turn off the placeholder text functionality.

        """
        self._placeholder_text = placeholder_text
        self._UpdatePlaceholderDisplay()

    def GetPlaceHolderText(self):
        """ Returns the placeholder text for this control.

        """
        return self._placeholder_text

    def ChangeValue(self, text):
        """ Overridden method which moves the insertion point to the end
        of the field when changing the text value. This causes the field
        to behave like Qt.

        """
        super(wxLineEdit, self).ChangeValue(text)
        self.SetInsertionPoint(len(text))

    def GetValue(self):
        """ Returns string value in the control, or an empty string if
        the placeholder text is active.

        """
        if self._placeholder_active:
            return ''
        return super(wxLineEdit, self).GetValue()

    def SetForegroundColour(self, wxColor, force=False):
        """ Sets the foreground color of the field. If the placeholder
        text is being shown, `force` must be True in order to override
        the placeholder text color.

        """
        self._user_fgcolor = wxColor
        if self._placeholder_active and not force:
            return
        super(wxLineEdit, self).SetForegroundColour(wxColor)

    def Clone(self, parent, style=None):
        """ Clone the wxLineEdit widget.

        Arguments
        ---------
        parent: wx.Window
            The parent to use for the new widget.

        style: int
            The style to use for the new widget. Default is to use the
            same style as the current object.

        """
        if style is None:
            style = self.GetWindowStyle()
        new_widget = self.__class__(parent=parent, style=style)
        new_widget._placeholder_text = self._placeholder_text
        new_widget._placeholder_active = self._placeholder_active
        new_widget.SetForegroundColour(self._user_fgcolor)
        new_widget.ChangeValue(self.GetValue())
        new_widget.SetInsertionPoint(self.GetInsertionPoint())
        new_widget.SetSelection(*self.GetSelection())
        # We have to clone the validator or WX will segfault when we
        # set it on the new widget.
        validator = self.GetValidator().Clone()
        if validator is not None:
            new_widget.SetValidator(validator)
        return new_widget


class WXField(WXControl, AbstractTkField):
    """ A wxPython implementation of a LineEdit.

    The LineEdit uses a custom wx.TextCtrl and provides a single line of
    editable text.

    """
    #--------------------------------------------------------------------------
    # Setup Methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying wxLineEdit.

        """
        # We have to do a bit of initialization in the create method
        # since wx requires the style of certain things to be set at
        # the point of instantiation
        style = wx.TE_PROCESS_ENTER
        shell = self.shell_obj

        if shell.read_only:
            style |= wx.TE_READONLY
        else:
            style &= ~wx.TE_READONLY
        
        password_mode = shell.password_mode
        if password_mode == 'normal':
            style &= ~wx.TE_PASSWORD
        elif password_mode == 'password':
            style |= wx.TE_PASSWORD
        else:
            style |= wx.TE_PASSWORD

        self.widget = wxLineEdit(parent=parent, style=style)

    def initialize(self):
        """ Initializes the attributes of the wxLineEdit.

        """
        super(WXField, self).initialize()
        shell = self.shell_obj
        self.set_validator(shell.validator)
        self.set_text(shell.validator.format(shell.value))
        self.set_placeholder_text(shell.placeholder_text)
        self.set_cursor_position(shell.cursor_position)
        self.set_max_length(shell.max_length)

    def bind(self):
        """ Binds the event handlers for the wx.TextCtrl.

        """
        super(WXField, self).bind()
        widget = self.widget
        widget.Bind(wx.EVT_TEXT, self.on_text_edited)
        widget.Bind(wx.EVT_TEXT_ENTER, self.on_return_pressed)
        widget.Bind(wx.EVT_KILL_FOCUS, self.on_lost_focus)
        widget.Bind(wx.EVT_LEFT_UP, self.on_selection_changed)

    #--------------------------------------------------------------------------
    # Shell Object Change Handlers
    #--------------------------------------------------------------------------
    def shell_validator_changed(self, validator):
        """ The change handler for the 'validator' attribute on the 
        shell object.
        
        """
        self.set_validator(validator)

    def shell_max_length_changed(self, max_length):
        """ The change handler for the 'max_length' attribute on the
        shell.

        """
        self.set_max_length(max_length)

    def shell_read_only_changed(self, read_only):
        """ The change handler for the 'read_only' attribute on the
        shell.

        """
        self.set_read_only(read_only)

    def shell_cursor_position_changed(self, cursor_position):
        """ The change handler for the 'cursor_position' attribute on
        the shell.

        """
        if not guard.guarded(self, 'updating_selection'):
            self.set_cursor_position(cursor_position)

    def shell_placeholder_text_changed(self, placeholder_text):
        """ The change handler for the 'placeholder_text' attribute on
        the shell.

        """
        self.set_placeholder_text(placeholder_text)

    def shell_password_mode_changed(self, mode):
        """ The change handler for the 'password_mode' attribute on the
        shell object.

        """
        self.set_password_mode(mode)
    
    #--------------------------------------------------------------------------
    # Manipulation Methods 
    #--------------------------------------------------------------------------
    def set_selection(self, start, end):
        """ Sets the selection in the widget between the start and
        end positions, inclusive.

        """
        self.widget.SetSelection(start, end)
        self._update_shell_selection_and_cursor()

    def select_all(self):
        """ Select all the text in the line edit.

        If there is no text in the line edit, the selection will be
        empty.

        """
        self.widget.SetSelection(-1, -1)
        self._update_shell_selection_and_cursor()

    def deselect(self):
        """ Deselect any selected text.

        Sets a selection with start == stop to deselect the current
        selection. The cursor is placed at the beginning of selection.

        """
        widget = self.widget
        start, end = widget.GetSelection()
        widget.SetSelection(start, start)
        self._update_shell_selection_and_cursor()

    def clear(self):
        """ Clear the line edit of all text.

        """
        self.widget.Clear()
        self._update_shell_selection_and_cursor()

    def backspace(self):
        """ Simple backspace functionality.

        If no text is selected, deletes the character to the left
        of the cursor. Otherwise, it deletes the selected text.

        """
        widget = self.widget
        start, end = widget.GetSelection()
        if start == end:
            start = end - 1
        widget.Remove(start, end)
        self._update_shell_selection_and_cursor()

    def delete(self):
        """ Simple delete functionality.

        If no text is selected, deletes the character to the right
        of the cursor. Otherwise, it deletes the selected text.

        """
        widget = self.widget
        start, end = widget.GetSelection()
        if start == end:
            end = end + 1
        widget.Remove(start, end)
        self._update_shell_selection_and_cursor()

    def end(self, mark=False):
        """ Moves the cursor to the end of the line.

        Arguments
        ---------
        mark : bool, optional
            If True, select the text from the current position to the end of
            the line edit. Defaults to False.

        """
        widget = self.widget
        if mark:
            start = widget.GetInsertionPoint()
            end = widget.GetLastPosition()
            widget.SetSelection(start, end)
        else:
            widget.SetInsertionPointEnd()
        self._update_shell_selection_and_cursor()

    def home(self, mark=False):
        """ Moves the cursor to the beginning of the line.

        Arguments
        ---------
        mark : bool, optional
            If True, select the text from the current position to
            the beginning of the line edit. Defaults to False.

        """
        widget = self.widget
        if mark:
            start = 0
            end = widget.GetInsertionPoint()
            widget.SetSelection(start, end)
        else:
            widget.SetInsertionPoint(0)
        self._update_shell_selection_and_cursor()

    def cut(self):
        """ Cuts the selected text from the line edit.

        Copies the selected text to the clipboard then deletes the selected
        text from the line edit.

        """
        self.widget.Cut()
        self._update_shell_selection_and_cursor()

    def copy(self): 
        """ Copies the selected text to the clipboard.

        """
        self.widget.Copy()
        self._update_shell_selection_and_cursor()

    def paste(self):
        """ Paste the contents of the clipboard into the line edit.

        Inserts the contents of the clipboard into the line edit at
        the current cursor position, replacing any selected text.

        """
        self.widget.Paste()
        self._update_shell_selection_and_cursor()

    def insert(self, text):
        """ Insert the text into the line edit.

        Inserts the given text at the current cursor position,
        replacing any selected text.

        Arguments
        ---------
        text : str
            The text to insert into the line edit.

        """
        self.widget.WriteText(text)
        self._update_shell_selection_and_cursor()

    def undo(self):
        """ Undoes the last operation.

        """
        self.widget.Undo()
        self._update_shell_selection_and_cursor()

    def redo(self):
        """ Redoes the last operation

        """
        self.widget.Redo()
        self._update_shell_selection_and_cursor()
    
    #--------------------------------------------------------------------------
    # Event handlers
    #--------------------------------------------------------------------------
    def on_text_edited(self, event):
        """ The event handler for when the user edits the text through 
        the ui.

        """
        event.Skip()
        self.shell_obj._field_text_edited()

    def on_return_pressed(self, event):
        """ The event handler for the return pressed event.

        """
        # Don't skip the event here or it will result in an annoying
        # Windows system beep.
        self.shell_obj._field_return_pressed()

    def on_lost_focus(self, event):
        """ The signal handler for the lost focus event.

        """
        event.Skip()
        self.shell_obj._field_lost_focus()

    def on_selection_changed(self, event):
        """ The event handler for a selection event.

        """
        event.Skip()
        self._update_shell_selection_and_cursor()

    #--------------------------------------------------------------------------
    # Updated Methods
    #--------------------------------------------------------------------------
    def get_text(self):
        """ Returns the current unicode text in the control.

        """
        return self.widget.GetValue()
    
    def set_text(self, text):
        """ Updates the text control with the given unicode text.

        """
        self.widget.ChangeValue(text)

    def set_validator(self, validator):
        """ Wraps the given Enaml validator in a custom wx.PyValidator 
        instance and applies it to the underlying control.

        """
        wxvalidator = WXFieldValidator(validator)
        self.widget.SetValidator(wxvalidator)
    
    def set_max_length(self, max_length):
        """ Set the max length of the control to max_length. If the max 
        length is <= 0 or > 32767 then the control will be set to hold 
        32kb of text.

        """
        if (max_length <= 0) or (max_length > 32767):
            max_length = 32767
        self.widget.SetMaxLength(max_length)

    def set_read_only(self, read_only):
        """ Sets the read only state of the widget. 

        """
        # Since the read-only state can only be set at creation time,
        # we need to create a new widget and swap it out. This is a
        # terrible hack that is forced upon us by wx limitations.
        style = self.widget.GetWindowStyle()
        if read_only:
            style |= wx.TE_READONLY
        else:
            style &= ~wx.TE_READONLY
        style |= wx.TE_PROCESS_ENTER
        self._update_widget_style(style)

    def set_placeholder_text(self, placeholder_text):
        """ Sets the placeholder text in the widget.

        """
        self.widget.SetPlaceHolderText(placeholder_text)

    def set_cursor_position(self, cursor_position):
        """ Sets the cursor position of the widget.

        """
        self.widget.SetInsertionPoint(cursor_position)

    def set_password_mode(self, password_mode):
        """ Sets the password mode of the wiget. Currently WXField only
        supports `password` and normal`.

        """
        # Since the read-only state can only be set at creation time,
        # we need to create a new widget and swap it out. This is a
        # terrible hack that is forced upon us by wx limitations.
        widget = self.widget
        style = widget.GetWindowStyle()
        if password_mode == 'normal':
            style &= ~wx.TE_PASSWORD
        elif password_mode == 'password':
            style |= wx.TE_PASSWORD
        else:
            style |= wx.TE_PASSWORD
        style |= wx.TE_PROCESS_ENTER
        self._update_widget_style(style)

    #--------------------------------------------------------------------------
    # Helper Methods
    #--------------------------------------------------------------------------
    def _update_shell_selection_and_cursor(self):
        """ Update the selection and cursor position for the shell
        object.

        """
        # This isn't nearly as nice as Qt. We can only update the 
        # selection due to a left up event because wx doesn't provide
        # any better/more relevant events. This means this function 
        # is just call whenver any change in the control *may* have
        # caused these things to update. We are doing way more work
        # here that we should be needing to, but the only way around
        # this that I see is to implement our own control from scratch.
        with guard(self, 'updating_selection'):
            self.shell_obj.selected_text = self.widget.GetStringSelection()
            self.shell_obj.cursor_position = self.widget.GetInsertionPoint()

    def _update_widget_style(self, style):
        """ Create a new CustomTextWidget widget with the given style
        and use it as the new widget, destroying the old one.

        """
        widget = self.widget
        new_widget = widget.Clone(parent=widget.GetParent(), style=style)
        widget.Destroy()
        self.widget = new_widget
        self.bind()
        new_widget.Show()
        # We need a relayout to put the new widget in the proper position.
        self.shell_obj.request_relayout()
    
    #--------------------------------------------------------------------------
    # Overrides
    #--------------------------------------------------------------------------
    def size_hint(self):
        """ Overridden size_hint method to add 44 pixels in width to 
        the field. This makes Wx consistent with Qt.

        """
        w, h = super(WXField, self).size_hint()
        return (w + 44, h)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt import QtCore, QtGui
from .qt_toggle_control import QtToggleControl

from ...components.toggle_button import AbstractTkToggleButton


class QtToggleButton(QtToggleControl, AbstractTkToggleButton):
    """ A Qt implementation of ToggleButton.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying QPushButton control which has the 
        'checkable' property set to True.

        """
        self.widget = QtGui.QPushButton(parent)
        # The QPushButton is not toggleable until it is 'setCheckable'. 
        # It must be toggleable before 'QtToggleControl.initialize' is 
        # called, so we do that here.
        self.widget.setCheckable(True)

    def initialize(self):
        """ Intializes the widget with the attributes of this instance.

        """
        super(QtToggleButton, self).initialize()
        self.set_icon(self.shell_obj.icon)
        self.set_icon_size(self.shell_obj.icon_size)
        
    def bind(self):
        """ Binds the event handlers for the toggle button.

        """
        super(QtToggleButton, self).bind()
        widget = self.widget
        widget.toggled.connect(self.on_toggled)
        widget.pressed.connect(self.on_pressed)
        widget.released.connect(self.on_released)
        
    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_icon_changed(self, icon):
        """ The change handler for the 'icon' attribute.

        """
        self.set_icon(icon)
        # If the icon of the button changes, the size hint has likely
        # changed and the layout system needs to be informed.
        self.shell_obj.size_hint_updated()

    def shell_icon_size_changed(self, icon_size):
        """ The change handler for the 'icon_size' attribute.

        """
        self.set_icon_size(icon_size)
        # If the icon of the button changes, the size hint has likely
        # changed and the layout system needs to be informed.
        self.shell_obj.size_hint_updated()

    def set_icon(self, icon):
        """ Sets the icon on the button control.

        """
        if icon is None:
            qicon = QtGui.QIcon()
        else:
            qicon = icon.as_QIcon()
        self.widget.setIcon(qicon)

    def set_icon_size(self, icon_size):
        """ Sets the icon size on the button control.

        """
        if (icon_size is not None and 
            icon_size.width > 0 and 
            icon_size.height > 0):
            self.widget.setIconSize(QtCore.QSize(*icon_size))




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import wx

from .wx_slider import WXSlider

from ...components.float_slider import AbstractTkFloatSlider


class WXFloatSlider(WXSlider, AbstractTkFloatSlider):
    """ A wxPython implementation of FloatSlider.

    """
    #--------------------------------------------------------------------------
    # Setup Methods
    #--------------------------------------------------------------------------
    def initialize(self):
        """ Initializes the attributes of the toolkit widget.

        """
        # Setting double buffered reduces flicker on Windows
        self.widget.SetDoubleBuffered(True)
        shell = self.shell_obj
        shell._down = False
        self.set_precision(shell.precision)
        super(WXFloatSlider, self).initialize()

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------

    def shell_precision_changed(self, precision):
        """ The change handler for the 'precision' attribute on the shell
        object.
        
        """
        self.set_precision(precision)

    #--------------------------------------------------------------------------
    # Event Handlers 
    #--------------------------------------------------------------------------
    def _on_slider_moved(self, event):
        """ The event handler for a slider drag event.

        """
        shell = self.shell_obj
        value = self.int_to_float(self.widget.GetValue())

        # Update the value only if tracking is turned on.
        if shell.tracking:
            shell.value = value
        
        # Since this event is only fired when the drag handle is 
        # moved with the mouse, it's safe to fire the move event
        # with no further checks.
        shell.moved(value)

        event.Skip()

    def _on_slider_keyed(self, event):
        """ The event handler for the user changing the value of the 
        slider through a key.

        """
        self.shell_obj.value = self.int_to_float(self.widget.GetValue())
        event.Skip()

    def _on_left_down(self, event):
        """ The event handler for the left mouse button being pressed.
        This estimates the position of the drag handle and then checks 
        if the mouse was pressed over it and then fires the `pressed` 
        event and sets the `down` attribute.

        """
        # Wx does not provide an event for detecting a mouse press
        # on the drag handle. So we have to estimate it by doing a
        # bit of geometry. This is not very accurate but probably
        # sufficient for most use cases.
        shell = self.shell_obj
        widget = self.widget
        mouse_position = event.GetPosition()
        slider_position = widget.GetValue()
        thumb = widget.GetThumbLength() / 2.0
        width, height = widget.GetClientSizeTuple()

        if widget.HasFlag(wx.SL_VERTICAL):
            position = mouse_position[1] / float(height)
        else:
            position = mouse_position[0] / float(width)

        slider_length = float(widget.GetMax() - widget.GetMin()) + 1.0
        minimum = (slider_position - thumb) / slider_length
        maximum = (slider_position + thumb) / slider_length

        hit = minimum <= position <= maximum

        if hit:
            shell._down = True
            shell.pressed()
        
        event.Skip()

    def _on_thumb_released(self, event):
        """ The event handler for drag handle being released. It
        updates the value in the shell and fires the appropriate
        events.

        """
        shell = self.shell_obj
        value = self.int_to_float(self.widget.GetValue())
        shell.value = value
        if shell._down:
            shell._down = False
            shell.released()
        event.Skip()

    #--------------------------------------------------------------------------
    # Widget Update Methods
    #--------------------------------------------------------------------------
    def set_precision(self, precision):
        """ Set the precision attribute in the Qt widget.

        """
        self.widget.SetRange(0, precision)

    def set_range(self, minimum, maximum):
        """ Set the range of the slider widget.

        Overridden to ignore the float values. The Qt widget knows nothing about
        them.

        """
        pass

    def set_position(self, value):
        """ Validate the position value.

        """
        self.widget.SetValue(self.float_to_int(value))

    def float_to_int(self, value):
        """ Convert a float value to an int value.

        """
        shell = self.shell_obj
        u = (value - shell.minimum) / (shell.maximum - shell.minimum)
        i = int(round(u * shell.precision))
        return i

    def int_to_float(self, value):
        """ Convert an int value to a float value.

        """
        shell = self.shell_obj
        u = float(value) / shell.precision
        x = u * shell.maximum + (1.0-u) * shell.minimum
        return x





#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import HasTraits, Float


class Myclass(HasTraits):
    """ This is a example class

    It is used to demonstrate the proposed guidelines for documenting traits
    based classes in an "sphinx-friendly" way.

    The traits are documented close to their definition by using a special
    comment ``#:`` prefix.

    """

    #: The x Float trait (default = 150.0)
    x = Float(150.0)

    # This is a comment autodoc ignores it
    #: The y Float trait (default = 0.0)
    y = Float(0.0)


class Otherclass(HasTraits):
    """ This is another example class using traits

    It is used to demonstrate the alternative method for documenting traits
    based classes in an "sphinx-friendly" way.

    The traits are documented at the start of the class definition.

    Attributes
    ----------
    x : Float, default = 150.0
        The x Float trait

    y : Float, default = 150.0
        The y Float trait

    """

    x = Float(150.0)

    # This is a comment autodoc ignores it
    y = Float(0.0)



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from datetime import date

from traits.api import TraitError

from .enaml_test_case import EnamlTestCase, required_method


class TestDateEdit(EnamlTestCase):
    """ Logic for testing the date edit components.

    Tooklit testcases need to provide the following functions

    Abstract Methods
    ----------------
    get_date(self, widget)
        Get the toolkits widget's active date.

    get_min_date(self, widget)
        Get the toolkits widget's minimum date attribute.

    get_max_date(self, widget)
        Get the toolkits widget's maximum date attribute.

    change_date(self, widget, date)
        Simulate a change date action at the toolkit widget.

    get_date_as_string(self, widget)
        Get the toolkits widget's active date as a string.

    """

    def setUp(self):
        """ Set up before the date_edit tests

        """
        enaml_source = """
enamldef MainView(MainWindow):
    attr events
    DateEdit:
        name = 'test'
        date_changed :: events.append(('date_changed', event.new))
"""
        self.events = []
        self.view = self.parse_and_create(enaml_source, events=self.events)
        self.component = self.component_by_name(self.view, 'test')
        self.widget = self.component.toolkit_widget

    def test_initialization_values(self):
        """ Test the initial attributes of the date edit component.

        """
        component = self.component
        self.assertEnamlInSync(component, 'date', date.today())
        self.assertEnamlInSync(component, 'min_date', date(1752, 9, 14))
        self.assertEnamlInSync(component, 'max_date', date(7999, 12, 31))
        self.assertEqual(self.events, [])

    def test_change_min_date(self):
        """ Test changing the minimum date.

        """
        component = self.component
        new_minimum = date(2000,1,1)
        component.min_date = new_minimum
        self.assertEnamlInSync(component, 'min_date', new_minimum)

    def test_change_max_date(self):
        """ Test changing the maximum date.

        """
        component = self.component
        new_maximum = date(2005,1,1)
        component.max_date = new_maximum
        self.assertEnamlInSync(component, 'max_date', new_maximum)

    def test_change_date_in_enaml(self):
        """ Test changing the current date through the component.

        """
        component = self.component
        new_date = date(2007,10,9)
        component.date = new_date
        self.assertEnamlInSync(component, 'date', new_date)
        self.assertEqual(self.events, [])

    def test_change_date_in_ui(self):
        """ Test changing the current date through the ui.

        """
        component = self.component
        widget = self.widget
        new_date = date(2007,10,9)
        self.change_date(widget, new_date)
        self.assertEnamlInSync(component, 'date', new_date)
        self.assertEqual(self.events, [('date_changed', new_date)])

    def test_invalid_min_date(self):
        """ Test changing to an invalid date below the min range.

        """
        component = self.component
        min_date = date(2000,2,3)
        component.min_date = min_date
        with self.assertRaises(TraitError):
            component.date = date(2000,1,1)
        self.assertEnamlInSync(component, 'date', date.today())
        self.assertEqual(self.events, [])

    def test_invalid_max_date(self):
        """ Test changing to an invalid date above the max range.

        """
        component = self.component
        component.date = date(2011,10,9)
        self.assertEqual(self.events, [])
        max_date = date(2014,2,3)
        component.max_date = max_date
        with self.assertRaises(TraitError):
            component.date = date(2016,10,9)
        self.assertEnamlInSync(component, 'date', date(2011,10,9))
        self.assertEqual(self.events, [])

    def test_change_maximum_and_date(self):
        """ Test setting maximum while the date is out of range.

        """
        component = self.component
        component.date = date(2007,10,9)
        component.max_date = date(2006,5,9)
        self.assertEnamlInSync(component, 'date', date(2006,5,9))
        self.assertEqual(self.events, [])

    def test_change_minimum_and_date(self):
        """ Test setting minimum while the date is out of range.

        """
        component = self.component
        component.date = date(2007,10,9)
        component.min_date = date(2010,5,9)
        self.assertEnamlInSync(component, 'date', date(2010,5,9))
        self.assertEqual(self.events, [])

    def test_change_range_invalid(self):
        """ Test setting minimum > maximum.

        """
        component = self.component
        component.min_date = date(2010,5,9)
        with self.assertRaises(TraitError):
            component.max_date = date(2006,5,9)

        component.max_date = date(2034,12,10)
        with self.assertRaises(TraitError):
            component.min_date = date(2034,12,14)

    def test_set_format(self):
        """ Test setting the output format

        """
        component = self.component
        widget = self.widget
        test_date = date(2007,10,9)
        component.date_format = 'MMM dd yyyy'
        component.date = date(2007,10,9)
        widget_string = self.get_date_as_string(widget)
        formated_date = unicode(test_date.strftime('%b %d %Y'), encoding='utf-8')
        self.assertEqual(widget_string, formated_date)

    #--------------------------------------------------------------------------
    # Special initialization tests
    #--------------------------------------------------------------------------
    def test_initial_too_late(self):
        """ Check initialization with an invalid late date.

        """
        enaml_source = """
import datetime
enamldef MainView(MainWindow):
    attr events
    DateEdit:
        name = 'test'
        date = datetime.date(2010, 1, 1)
        min_date = datetime.date(1990, 1, 1)
        max_date = datetime.date(2000, 1, 1)
        date_changed :: events.append('date_changed')
"""
        events = []
        # FIXME: need make a more refined check, this is not the best way
        with self.assertRaises(TraitError):
            self.parse_and_create(enaml_source, events=events)

    def test_initial_too_early(self):
        """ Check initialization with an invalid early date.

        """
        enaml_source = """
import datetime
enamldef MainView(MainWindow):
    attr events
    DateEdit:
        name = 'test'
        min_date = datetime.date(1990, 1, 1)
        max_date = datetime.date(2000, 1, 1)
        date = datetime.date(1980, 1, 1)
        date_changed :: events.append('date_changed')
"""
        events = []
        # FIXME: need make a more refined check, this is not the best way
        with self.assertRaises(TraitError):
            self.parse_and_create(enaml_source, events=events)

    #--------------------------------------------------------------------------
    # Abstract methods
    #--------------------------------------------------------------------------
    @required_method
    def get_date(self, widget):
        """  Get the toolkits widget's active date.

        """
        pass

    @required_method
    def get_min_date(self, widget):
        """  Get the toolkits widget's maximum date attribute.

        """
        pass

    @required_method
    def get_max_date(self, widget):
        """ Get the toolkits widget's minimum date attribute.

        """
        pass

    @required_method
    def change_date(self, widget, date):
        """ Simulate a change date action at the toolkit widget.

        """
        pass

    @required_method
    def get_date_as_string(self, widget):
        """  Get the toolkits widget's active date as a string.

        """
        pass




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import weakref

import wx
import wx.lib.newevent

from .wx_base_widget_component import WXBaseWidgetComponent

from ...components.action import AbstractTkAction


#: A custom event emitted by the wxEnamlAction when it has been triggered
wxEnamlActionTriggered, EVT_ACTION_TRIGGERED = wx.lib.newevent.NewEvent()

#: A custom event emitted by the wxEnamlAction when it has been toggled
wxEnamlActionToggled, EVT_ACTION_TOGGLED = wx.lib.newevent.NewEvent()


class wxEnamlAction(wx.EvtHandler):
    """ An wx.EvtHandler which *acts* like a wx equivalent to a QAction.
    It works by managing an internal wxMenuItem. Menus which use 
    this action must call the Install() method in order for the
    underlying wx.MenuItem to be created. This is due to the fact 
    that using the wx.MenuItem constructor from Python will lead 
    to segfaults when calling menu.AppendItem(item), hence creating
    a usable subclass of wx.MenuItem is out of the question.

    """
    def __init__(self, parent):
        """ Initialize a wxAction.

        Parameters
        ----------
        parent : wxEnamlMenu
            The wxEnamlMenu instance which is the parent of this action.

        """
        super(wxEnamlAction, self).__init__()

        # The parent wx.Menu of this action
        self._parent = weakref.ref(parent)

        # The internal wx.MenuItem that this action will manage
        self._menu_item = None

        # The text of the action
        self._text = ''

        # Whether or not the action is a separator
        self._separator = False

        # Whether or not the action is enabled
        self._enabled = True

        # Whether or not the action is checkable
        self._checkable = False
        
        # Whether or not the action is checked
        self._checked = False

        # The help text for the action
        self._help_text = ''

        # The icon for the action
        self._icon = None

        # Find the parent on which we need to bind the event handler. 
        # If this action is contained in a MenuBar, this will be a
        # wx.Frame, otherwise it will be a wx.Menu used as a sole 
        # popup menu.
        while parent is not None:
            last = parent
            parent = parent.GetParent()
        last.Bind(wx.EVT_MENU, self.OnMenuEvent)

        # Store away a weak reference to the object on which we
        # bound the event handler, so we can unbind it in the 
        # Destroy method.
        self._binder = weakref.ref(last)

    def OnMenuEvent(self, event):
        """ The event handler for menu events on this action.

        """
        menu_item = self._menu_item
        if menu_item is not None and menu_item.GetId() == event.GetId():
            # If the action is checkable, we first need to emit
            # the toggled event. This behavior is the same as Qt.
            if self._checkable:
                self._checked = menu_item.IsChecked()
                evt = wxEnamlActionToggled()
                wx.PostEvent(self, evt)
            # Next, we emit the standard triggered Event.
            evt = wxEnamlActionTriggered()
            wx.PostEvent(self, evt)
        else:
            # Only skip the event if it's not handled by this action.
            # Otherwise, we'll get multiple notifications.
            event.Skip()

    def GetParent(self):
        """ Returns the parent of the action or None if the parent has
        already been garbage collected.

        """
        return self._parent()

    def GetMenuItem(self):
        """ Returns the internal wx.MenuItem being managed by the action.

        """
        return self._menu_item
    
    def Install(self, idx=None):
        """ Adds this action to the internal parent wx.Menu object at 
        the given index. If the index is None, the item is appended to
        the end of the menu.

        """
        menu = self._parent()
        if menu is not None:
            # When creating a new menu item, we need to create a new
            # id for that item, as well as a valid name for the item.
            # If we don't give it a valid name (an empty string is 
            # invalid), then it will segfault. We create the item with 
            # a temporary name, and update it after the item has been 
            # created. We also fully update the rest of item state after 
            # it has been created.
            menu_id = wx.NewId()
            temp_name = str(menu_id)

            if idx is None:
                idx = menu.GetMenuItemCount()
            
            if self._separator:
                self._menu_item = menu.InsertSeparator(idx)
            elif self._checkable:
                self._menu_item = menu.InsertCheckItem(idx, menu_id, temp_name)
            else:
                self._menu_item = menu.Insert(idx, menu_id, temp_name)

            # Update the state of the menu item after it's been created.
            self.Enable(self._enabled)
            self.SetText(self._text)
            self.SetChecked(self._checked)
            self.SetHelp(self._help_text)
            self.UpdateMenuBitmaps()

    def GetText(self, text):
        """ Returns the current text for the action.

        """
        return self._text

    def SetText(self, text):
        """ Sets the text for the action. If the given text is the 
        empty string, then a unique mangled name will be generated
        for the action.

        """
        item = self._menu_item
        if item is not None:
            if not text:
                text = 'Action_' + str(item.GetId())
            item.SetText(text)
        else:
            if not text:
                text = 'Action_' + str(id(self))
        self._text = text
    
    def GetHelp(self):
        """ Returns the help text for the action.

        """
        return self._help_text
    
    def SetHelp(self, text):
        """ Sets the help text for the action.

        """
        self._help_text = text
        menu_item = self._menu_item
        if menu_item is not None:
            menu_item.SetHelp(text)

    def GetIcon(self):
        """ Return the WXIcon instance in use for this action.

        """
        return self._icon

    def SetIcon(self, icon):
        """ Set the icon for this action. The icon should be an instance
        of the Enaml WXIcon.

        """
        self._icon = icon

    def IsSeparator(self):
        """ Returns whether or not this action is a separator.

        """
        return self._separator

    def SetSeparator(self, separator):
        """ Sets whether or not this action is a separator.

        """
        # If the menu item has already been created, it needs
        # be removed from the parent and re-installed since 
        # the separator state cannot be changed in-place.
        self._separator = separator
        menu_item = self._menu_item
        if menu_item is not None:
            parent = self._parent()
            if parent is not None:
                idx = parent.GetMenuItems().index(menu_item)
                parent.Remove(menu_item.GetId())
                self.Install(idx)

    def IsCheckable(self):
        """ Returns whether or not the action is checkable.

        """
        return self._checkable
    
    def SetCheckable(self, checkable):
        """ Sets whether or not the action is checkable.

        """
        # If the menu item has already been created, it needs
        # be removed from the parent and re-installed since 
        # the checkable state cannot be changed in-place.
        self._checkable = checkable
        menu_item = self._menu_item
        if menu_item is not None:
            parent = self._parent()
            if parent is not None:
                idx = parent.GetMenuItems().index(menu_item)
                parent.Remove(menu_item.GetId())
                self.Install(idx)

    def IsChecked(self):
        """ Returns whether or not the action is checked. This only
        has semantic meaning if the action is also checkable.

        """
        return self._checked
    
    def SetChecked(self, checked):
        """ Sets whether or not the action is checked. This only has
        semantic meaning if the action is also checkable.

        """
        self._checked = checked
        menu_item = self._menu_item
        if menu_item is not None and self._checkable:
            menu_item.Check(checked)

    def Enable(self, enabled):
        """ Enables or disables the action.

        """
        self._enabled = enabled
        menu_item = self._menu_item
        if menu_item is not None:
            menu_item.Enable(enabled)
            self.UpdateMenuBitmaps()
    
    def UpdateMenuBitmaps(self):
        """ Update the menu item bitmaps correctly for the current state.
        
        """
        icon = self._icon
        menu_item = self._menu_item
        if menu_item is None:
            return
        if icon is not None:
            mode = 'normal' if self._enabled else 'disabled'
            size = (16, 16)
            if self.IsCheckable():
                checked_img = icon.get_image(size, mode=mode, state='on')
                checked_bmp = checked_img.as_wxBitmap()
                unchecked_img = icon.get_image(size, mode=mode, state='off')
                unchecked_bmp = unchecked_img.as_wxBitmap()
                menu_item.SetBitmaps(checked_bmp, unchecked_bmp)
            else:
                img = icon.get_image(size, mode=mode)
                bmp = img.as_wxBitmap()
                menu_item.SetBitmaps(bmp, wx.NullBitmap)
        else:
            menu_item.SetBitmaps(wx.NullBitmap, wx.NullBitmap)
    
    def Destroy(self):
        """ A destructor method which removes the menu item from its
        parent before destroying the item and itself.

        """
        if self:
            menu_item = self._menu_item
            parent = self._parent()
            if menu_item is not None and parent is not None:
                try:
                    # This will raise a ValueError if the menu item is
                    # not in the list. Looping over the list and checking
                    # for equality won't work; a wx-ism
                    parent.GetMenuItems().index(menu_item)
                except ValueError:
                    pass
                else:
                    parent.Remove(menu_item.GetId())

            if menu_item:
                menu_item.Destroy()
            self.menu_item = None

            # Before destroying the action, we need to unbind the event
            # handler, or it will continue being called for every 
            # menu event which will result in PyDeadObject errors.
            binder = self._binder()
            if binder is not None:
                binder.Unbind(wx.EVT_MENU, handler=self.OnMenuEvent)
            super(wxEnamlAction, self).Destroy()

    def Freeze(self):
        """ An api compatibility method. This is a no-op since actions
        cannot be frozen or thawed.

        """
        pass

    def Thaw(self):
        """ An api compatibility method. This is a no-op since actions
        cannot be frozen or thawed.

        """
        pass
        

class WXAction(WXBaseWidgetComponent, AbstractTkAction):
    """ A wx implementation of Action.

    """
    #--------------------------------------------------------------------------
    # Setup Methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Creates the underlying QAction object.

        """
        self.widget = wxEnamlAction(parent)

    def initialize(self):
        """ Initializes the underlying QAction object.

        """
        super(WXAction, self).initialize()
        shell = self.shell_obj
        self.set_enabled(shell.enabled)
        self.set_text(shell.text)
        self.set_checkable(shell.checkable)
        self.set_checked(shell.checked)
        self.set_status_tip(shell.status_tip)
        self.set_tool_tip(shell.tool_tip)
        self.set_icon(shell.icon)
        self.set_description(shell.description)
        self.set_separator(shell.separator)
        
    def bind(self):
        """ Binds the event handlers for the underlying QAction object.

        """
        super(WXAction, self).bind()
        self.widget.Bind(EVT_ACTION_TRIGGERED, self.on_triggered)
        self.widget.Bind(EVT_ACTION_TOGGLED, self.on_toggled)

    #--------------------------------------------------------------------------
    # Change Handlers 
    #--------------------------------------------------------------------------
    def shell_enabled_changed(self, enabled):
        """ The change handler for the 'enabled' attribute of the shell
        object. Sets the widget enabled according to the given boolean.

        """
        self.set_enabled(enabled)

    def shell_text_changed(self, text):
        """ The change handler for the 'text' attribute on the shell
        object.

        """
        self.set_text(text)
    
    def shell_checkable_changed(self, checkable):
        """ The change handler for the 'checkable' attribute on the 
        shell object.

        """
        self.set_checkable(checkable)
    
    def shell_checked_changed(self, checked):
        """ The change handler for the 'checked' attribute on the
        shell object.

        """
        self.set_checked(checked)

    def shell_status_tip_changed(self, status_tip):
        """ The change handler for the 'status_tip' attribute on the 
        shell object.

        """
        self.set_status_tip(status_tip)
    
    def shell_tool_tip_changed(self, tool_tip):
        """ The change handler for the 'tool_tip' attribute on the 
        shell object.

        """
        self.set_tool_tip(tool_tip)
    
    def shell_icon_changed(self, icon):
        """ The change handler for the 'icon' attribute on the shell 
        object.

        """
        self.set_icon(icon)
    
    def shell_description_changed(self, description):
        """ The change handler for the 'description' attribute on the 
        shell object.

        """
        self.set_description(description)

    def shell_separator_changed(self, separator):
        """ The change handler for the 'separator' attribute on the
        shell object.

        """
        self.set_separator(separator)

    #--------------------------------------------------------------------------
    # Signal Handlers
    #--------------------------------------------------------------------------
    def on_triggered(self, event):
        """ A signal handler for the 'triggered' signal of the QAction.

        """
        self.shell_obj.triggered()

    def on_toggled(self, event):
        """ A signal handler for the 'toggled' signal of the QAction.

        """
        shell = self.shell_obj
        shell.checked = self.widget.IsChecked()
        shell.toggled()

    #--------------------------------------------------------------------------
    # Widget Update Methods 
    #--------------------------------------------------------------------------
    def set_enabled(self, enabled):
        """ Sets the enabled state of the action.

        """
        self.widget.Enable(enabled)

    def set_text(self, text):
        """ Sets the text of the action.

        """
        self.widget.SetText(text)

    def set_checkable(self, checkable):
        """ Sets whether or not the action is checkable.

        """
        self.widget.SetCheckable(checkable)
    
    def set_checked(self, checked):
        """ Sets whether or not the action is checked.

        """
        self.widget.SetChecked(checked)

    def set_status_tip(self, status_tip):
        """ Sets the status tip text for the action.

        """
        self.widget.SetHelp(status_tip)
    
    def set_tool_tip(self, tool_tip):
        """ Sets the tool tip text for the action.

        """
        # XXX not yet supported
    
    def set_icon(self, icon):
        """ Sets the icon for the action.

        """
        self.widget.SetIcon(icon)
    
    def set_description(self, description):
        """ Sets the description text for the action.

        """
        # XXX not yet supported

    def set_separator(self, separator):
        """ Sets whether or not the QAction is a separator.

        """
        self.widget.SetSeparator(separator)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import abstractmethod

from traits.api import Str, Any, Instance

from .container import Container, AbstractTkContainer


class AbstractTkTab(AbstractTkContainer):
    """ The abstract toolkit Tab interface.

    """
    @abstractmethod
    def shell_title_changed(self, title):
        """ The change handler for the 'title' attribute on the shell
        object.

        """
        raise NotImplementedError
    
    @abstractmethod
    def shell_icon_changed(self, icon):
        """ The change handler for the 'icon' attribute on the shell
        object.

        """
        raise NotImplementedError


class Tab(Container):
    """ A Container subclass that represents a tab in a Tabbed container.
    
    """
    #: The title of this tab in the Tabbed container.
    title = Str

    #: The icon displayed on this tab, if any.
    # XXX handle icons!
    icon = Any

    #: Overridden parent class trait
    abstract_obj = Instance(AbstractTkTab)

    def set_visible(self, foo):
        """ Overridden visibility method which turns off the ability
        to change the visibility of a tab. A Tab's visibility is 
        controlled implicitly by the TabGroup and the currently
        selected tab.

        """
        pass
    



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import datetime
import locale
import sys
import time

from .abstract_locale import AbstractLocale


# nl_langinfo is not available on Windows and as a result locale is 
# mostly useless. So we have to fake a bunch of stuff for that platform
# and just hard-code it to the english equivalents. 
_win_defaults = {
    'am': 'AM',
    'pm': 'PM',
    'months': {
        1: 'January',
        2: 'February',
        3: 'March',
        4: 'April',
        5: 'May',
        6: 'June',
        7: 'July',
        8: 'August',
        9: 'September',
        10: 'October',
        11: 'November',
        12: 'December',
    },
    'ab_months': {
        1: 'Jan',
        2: 'Feb',
        3: 'Mar',
        4: 'Apr',
        5: 'May',
        6: 'Jun',
        7: 'Jul',
        8: 'Aug',
        9: 'Sep',
        10: 'Oct',
        11: 'Nov',
        12: 'Dec',
    },
    'days': {
        1: 'Sunday',
        2: 'Monday',
        3: 'Tuesday',
        4: 'Wednesday',
        5: 'Thursday',
        6: 'Friday',
        7: 'Saturday',
    },
    'ab_days': {
        1: 'Sun',
        2: 'Mon',
        3: 'Tue',
        4: 'Wed',
        5: 'Thu',
        6: 'Fri',
        7: 'Sat',
    },
    'time': '%I:%M:%S %p',
    'ab_time': '%H:%M:%S',
    'date': '%m/%d/%Y',
    'datetime': '%a %b %e %X %Y',
}


def _convert_base(value, base):
    """ Convert a base10 integral number into a string represented in
    a different number base. Bases from 2-36 inclusive are supported.
    
    """
    if not isinstance(value, (int, long)):
        msg = 'Expected integral number type. Got %s instead.'
        raise TypeError(msg % type(value))
    
    if not isinstance(base, int):
        msg = 'Expected integer base. Got %d instead.'
        raise TypeError(msg % base)

    if base < 2 or base > 36:
        msg = 'Base must be between 2 and 36 inclusive. Got %d instead.'
        raise ValueError(msg % base)
    
    digits = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    parts = []
    push = parts.append
    
    prefix = False
    if value < 0:
        prefix = True
        value = abs(value)
    
    div = divmod
    while value != 0:
        value, mod = div(value, base)
        push(digits[mod])
    
    if prefix:
        push('-')

    return ''.join(reversed(parts))


class SystemLocale(AbstractLocale):
    """ A concrete implementation of AbstractLocale which uses the Python
    'locale' module for locale-related information. 

    The SystemLocale does not maintain any state and will always reflect 
    the current system locale information as given by the 'locale' module.

    """
    @classmethod
    def default(cls):
        """ Initializes the system locale if not already initialized and
        returns an instance of SystemLocale.

        """
        loc = locale.getlocale()
        if loc == (None, None):
            locale.setlocale(locale.LC_ALL, '')
        return cls()
    
    #--------------------------------------------------------------------------
    # Properties
    #--------------------------------------------------------------------------
    @property
    def name(self):
        """ The RFC 1766 language code for the locale.

        """
        return unicode(locale.getlocale()[0])

    @property
    def encoding(self):
        """ The encoding for strings in the locale.

        """
        return unicode(locale.getlocale()[1])
                
    @property
    def am_string(self):
        """ The string for "AM" time in the locale.

        """
        if sys.platform == 'win32':
            r = _win_defaults['am']
        else:
            r = locale.nl_langinfo(locale.AM_STR)
        return unicode(r)
    
    @property
    def pm_string(self):
        """ The string for "PM" time in the locale.

        """
        if sys.platform == 'win32':
            r = _win_defaults['pm']
        else:
            r = locale.nl_langinfo(locale.PM_STR)
        return unicode(r)

    @property
    def decimal_point(self):
        """ The decimal separator for the locale.

        """
        return unicode(locale.localeconv()['decimal_point'])
            
    @property
    def monetary_decimal_point(self):
        """ The decimal separator for monetary values for the locale.

        """
        return unicode(locale.localeconv()['mon_decimal_point'])

    @property
    def exponential(self):
        """ The exponential character for the locale.

        """
        # Symbol not provided by locale module, so use Python default
        return u'e'

    @property
    def thousands_separator(self):
        """ The thousands separator for the locale.

        """
        return unicode(locale.localeconv()['thousands_sep'])
    
    @property
    def monetary_thousands_separator(self):
        """ The thousands separator for monetary values for the locale.

        """
        return unicode(locale.localeconv()['mon_thousands_sep'])

    @property
    def negative_sign(self):
        """ The monetary negative sign for the locale.

        """
        return unicode(locale.localeconv()['negative_sign'])
                
    @property
    def positive_sign(self):
        """ The monetary positive sign for the locale.

        """
        return unicode(locale.localeconv()['positive_sign'])
            
    #--------------------------------------------------------------------------
    # Methods
    #--------------------------------------------------------------------------
    def month_name(self, month_int, abbreviated=False):
        """ The month name for the given month integer.

        Parameters
        ----------
        month_int : int
            An integer in the range 1 to 12 inclusive.
        
        abbreviated : bool, optional
            If True, return the abbreviate version of the month name.
            The default is False.
        
        Returns
        -------
        result : string
            The month name or an empty string in the month int is out
            of range.

        """
        if month_int < 1 or month_int > 12:
            return ''
        if sys.platform == 'win32':
            if abbreviated:
                key = 'ab_months'
            else:
                key = 'months'
            r = _win_defaults[key][month_int]
        else:
            prefix = 'AB' if abbreviated else ''
            attr = prefix + ('MON_%i' % month_int)
            flag = getattr(locale, attr)
            r = locale.nl_langinfo(flag)
        return unicode(r)

    def day_name(self, day_int, abbreviated=False):
        """ The day name for the given day integer.

        Parameters
        ----------
        day_int : int
            An integer in the range 1 to 7 inclusive.
        
        abbreviated : bool, optional
            If True, return the abbreviate version of the day name.
            The default is False.
        
        Returns
        -------
        result : string
            The day name or an empty string in the day int is out
            of range.

        """
        if day_int < 1 or day_int > 7:
            return ''
        if sys.platform == 'win32':
            if abbreviated:
                key = 'ab_days'
            else:
                key = 'days'
            r = _win_defaults[key][day_int]
        else:
            prefix = 'AB' if abbreviated else ''
            attr = prefix + ('DAY_%i' % day_int)
            flag = getattr(locale, attr)
            r = locale.nl_langinfo(flag)
        return unicode(r)
        
    def time_format(self, abbreviated=False):
        """ The format string for a time value.

        Parameters
        ----------
        abbreviated : bool, optional
            If True, return the abbreviate version of the time format.
            The defaults is False.

        Returns
        -------
        result : string
            The time format string.

        """
        if sys.platform == 'win32':
            if abbreviated:
                key = 'ab_time'
            else:
                key = 'time'
            r = _win_defaults[key]
        else:
            flag = locale.T_FMT if abbreviated else locale.T_FMT_AMPM
            r = locale.nl_langinfo(flag)
        return unicode(r)

    def date_format(self, abbreviated=False):
        """ The format string for a date value.

        Parameters
        ----------
        abbreviated : bool, optional
            If True, return the abbreviated version of the date format.
            The defaults is False.

        Returns
        -------
        result : string
            The date format string.
        
        Notes
        -----
        Python's builtin locale (based on langinfo) does not distinguish
        abbreviated and non-abbreviated date formats. So the keyword
        argument is ignored in this implementation.

        """
        if sys.platform == 'win32':
            r = _win_defaults['date']
        else:
            # Python doesn't support %e in strptime, but %d is equivalent
            r = locale.nl_langinfo(locale.D_FMT).replace('%e', '%d')
        return unicode(r)

    def datetime_format(self, abbreviated=False):
        """ The format string for a datetime value.

        Parameters
        ----------
        abbreviated : bool, optional
            If True, return the abbreviated version of the datetime
            format. The defaults is False.

        Returns
        -------
        result : string
            The datetime format string.

        Notes
        -----
        Python's builtin locale (based on langinfo) does not distinguish
        abbreviated and non-abbreviated datetime formats. So the keyword
        argument is ignored in this implementation.

        """
        if sys.platform == 'win32':
            r = _win_defaults['datetime']
        else:
            # Python doesn't support %e in strptime, but %d is equivalent
            r = locale.nl_langinfo(locale.D_T_FMT).replace('%e', '%d')
        return unicode(r)
        
    #--------------------------------------------------------------------------
    # Conversions
    #--------------------------------------------------------------------------
    def to_time(self, time_str, abbreviated=False):
        """ Parse the given time string into a datetime.time object.

        Parameters
        ----------
        time_str : string
            A string formatted according to time_format().

        abbreviated : bool, optional
            Whether or not the time string is using the abbreviated 
            format. The default is False.
        
        Returns
        -------
        result : datetime.time
            The datetime.time object for the string.

        """
        fmt = self.time_format(abbreviated)
        struct_time = time.strptime(time_str, fmt)
        return datetime.time(*struct_time[3:6])

    def to_date(self, date_str, abbreviated=False):
        """ Parse the date string into a datetime.date object.

        Parameters
        ----------
        date_str : string
            A string formatted according to date_format().

        abbreviated : bool, optional
            Whether or not the date string is using the abbreviated 
            format. The default is False.
        
        Returns
        -------
        result : datetime.date
            The datetime.date object for the string.

        """
        fmt = self.date_format(abbreviated)
        struct_time = time.strptime(date_str, fmt)
        return datetime.date(*struct_time[0:3])
    
    def to_datetime(self, datetime_str, abbreviated=False):
        """ Parse the datetime string into a datetime.datetime object.

        Parameters
        ----------
        datetime_str : string
            A string formatted according to datetime_format().

        abbreviated : bool, optional
            Whether or not the datetime string is using the abbreviated 
            format. The default is False.
        
        Returns
        -------
        result : datetime.datetime
            The datetime.datetime object for the string.

        """
        fmt = self.datetime_format(abbreviated)
        struct_time = time.strptime(datetime_str, fmt)
        return datetime.datetime(*struct_time[:6])

    def to_int(self, int_str, base=10):
        """ Parse the string into an integer using the given base.

        Parameters
        ----------
        int_str : string
            The string representing the integer.
        
        base : int, optional
            The number base to use when parsing the string. The valid
            values are the same as those passed to Python's builtin 
            'int'. The default is 10.
        
        Returns
        -------
        result : int
            The parsed integer.

        """
        func = lambda val: int(val, base)
        return locale.atof(int_str, func)
    
    def to_long(self, long_str, base=10):
        """ Parse the string into a long integer using the given base.

        Parameters
        ----------
        long_str : string
            The string representing the integer.
        
        base : int, optional
            The number base to use when parsing the string. The valid
            values are the same as those passed to Python's builtin 
            'int'. The default is 10.
        
        Returns
        -------
        result : long
            The parsed long integer.

        """ 
        func = lambda val: long(val, base)
        return locale.atof(long_str, func)

    def to_float(self, float_str):
        """ Parse the given string into a float.

        Parameters
        ----------
        float_str : string
            The string representing the float.
        
        Returns
        -------
        result : float
            The parsed float.

        """
        return locale.atof(float_str)

    def to_complex(self, complex_str):
        """ Parse the given string into a complex.

        Parameters
        ----------
        complex_str : string
            The string representing the complex.
        
        Returns
        -------
        result : complex
            The parsed complex.

        """
        text = complex_str.replace(self.thousands_separator, u'')
        text = text.replace(self.decimal_point, u'.')
        return complex(text)

    def to_currency(self, currency_str, symbol=True, group=False, intl=False):
        """ Parse the given currency string into a float.

        Parameters
        ----------
        currency_str : string
            The string representing the currency.
        
        symbol : bool, optional
            If True, the currency symbol is included in the string. The
            default is True.
        
        group : bool, optional
            If True, the grouping marks are included in the string. The
            default is False.
        
        intl : bool, optional
            If True, the international version of the currency symbol
            is used in the string. This is only used if 'symbol' is
            True. The default is False.

        Returns
        -------
        result : float
            The parsed float.

        """
        curr = currency_str
        conv = locale.localeconv()
        if symbol:
            if intl:
                curr = curr.replace(conv['int_curr_symbol'], '')
            else:
                curr = curr.replace(conv['currency_symbol'], '')
        if group:
            curr = curr.replace(conv['mon_thousands_sep'], '')
        curr = curr.replace(conv['mon_decimal_point'], '.').strip()
        return float(curr)

    def from_time(self, time_obj, abbreviated=False):
        """ Convert the given time object into a locale specific string.

        Parameters
        ----------
        time_obj : datetime.time
            The time object for conversion.
        
        abbreviated : bool, optional
            Whether or not the time string should use the abbreviated 
            format. The default is False.

        Returns
        -------
        result : unicode string
            The locale formatted time string.

        """
        return unicode(time_obj.strftime(self.time_format(abbreviated)))
    
    def from_date(self, date_obj, abbreviated=False):
        """ Convert the given date object into a locale specific string.

        Parameters
        ----------
        date_obj : datetime.date
            The date object for conversion.
        
        abbreviated : bool, optional
            Whether or not the date string should use the abbreviated 
            format. The default is False.

        Returns
        -------
        result : unicode string
            The locale formatted date string.

        """
        return unicode(date_obj.strftime(self.date_format(abbreviated)))
    
    def from_datetime(self, datetime_obj, abbreviated=False):
        """ Convert the given datetime object into a locale specific 
        string.

        Parameters
        ----------
        datetime_obj : datetime.datetime
            The datetime object for conversion.
        
        abbreviated : bool, optional
            Whether or not the datetime string should use the abbreviated 
            format. The default is False.

        Returns
        -------
        result : string
            The locale formatted datetime string.

        """
        s = datetime_obj.strftime(self.datetime_format(abbreviated))
        return unicode(s)

    def from_int(self, value, base=10, group=False):
        """ Format the integer value according to the locale.

        Parameters
        ----------
        value : int
            The integer to convert a string.
        
        base : int, optional
            The number base to use when parsing the string. The valid
            values are the same as those passed to Python's builtin 
            'int'. The default is 10.

        group : bool, optional
            Whether or not to add group separators in the result. The
            default is False.
        
        Returns
        -------
        result : unicode string
            The locale formatted integer string.
        
        """
        # We can specially format bases 2, 8, 16. The others we format
        # as normal integers after converting them to a proper base.
        if base == 2:
            r = bin(value)
        elif base == 8:
            r = locale.format_string('%#o', value, group)
        elif base == 10:
            r = locale.format_string('%d', value, group)
        elif base == 16:
            r = locale.format_string('%#X', value, group)
        else:
            r = _convert_base(value, base)
        return unicode(r)

    def from_long(self, value, base=10, group=False):
        """ Format the long integer value according to the locale.

        Parameters
        ----------
        value : long
            The long integer to convert a string.
        
        base : int, optional
            The number base to use when parsing the string. The valid
            values are the same as those passed to Python's builtin 
            'int'. The default is 10.

        group : bool, optional
            Whether or not to add group separators in the result. The
            default is False.
        
        Returns
        -------
        result : unicode string
            The locale formatted long integer string.
        
        """
        # The int conversion can handle longs
        return self.from_int(value, base, group)

    def from_float(self, value, group=False, prec=-1):
        """ Format the float value according to the locale.

        Parameters
        ----------
        value : float
            The float to convert a string.
        
        group : bool, optional
            Whether or not to add group separators in the result. The
            default is False.
        
        prec : int, optional
            The number of signficant digits to display in the string. 
            The default is -1 which means the decision is left up
            to the implementor.

        Returns
        -------
        result : unicode string
            The locale formatted integer string.
        
        """
        if prec < 0:
            s = '%g'
        else:
            s = '%' + '.%dg' % prec
        return unicode(locale.format_string(s, value, group))

    def from_complex(self, value, group=False, prec=-1):
        """ Format the complex value according to the locale.

        Parameters
        ----------
        value : complex
            The complex to convert a string.
        
        group : bool, optional
            Whether or not to add group separators in the result. The
            default is False.
        
        prec : int, optional
            The number of signficant digits to display for each part. 
            The default is -1 which means the decision is left up
            to the implementor.

        Returns
        -------
        result : unicode string
            The locale formatted complex string.
        
        """
        real = self.from_float(value.real, group, prec)
        imag = self.from_float(value.imag, group, prec)
        if imag.startswith('-'):
            res = real + imag + 'j'
        else:
            res = real + '+' + imag + 'j'
        return res

    def from_currency(self, value, symbol=True, group=False, intl=False):
        """ Convert the given value into a locale specific currency
        string.

        Parameters
        ----------
        value : int, long, or float
            The value representing the currency
        
        symbol : bool, optional
            If True, the currency symbol is included in the string. The
            default is True.
        
        group : bool, optional
            If True, the grouping marks are included in the string. The
            default is False.
        
        intl : bool, optional
            If True, the international version of the currency symbol
            is used in the string. This is only used if 'symbol' is
            True. The default is False.
            
        Returns
        -------
        result : string
            The locale formatted currency string.
             
        """
        return unicode(locale.currency(value, symbol, group, intl))




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
""" The DataSource api for providing live streaming data to a Chaco plot.

"""
from abc import ABCMeta, abstractmethod
from enaml.core.signaling import Signal
import numpy as np


class AbstractDataSource(object):
    """ An abstract base class which defines the interface for a data
    source to be used by a live plot.

    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def subscribe(self, callback):
        """ Subscribe a callback to be called by the data source when
        new data is available. The argument will be a structured numpy
        array representing all of the data. Only a weak reference is
        maintained to the callback.

        """
        raise NotImplementedError

    @abstractmethod
    def unsubscribe(self, callback):
        """ Unsubscribe a callback that was registered previously 
        via a call to subscribe.

        """
        raise NotImplementedError

    @abstractmethod
    def bind(self):
        """ Bind this data source to the publisher. The data source
        will start listening to the publisher immediately.

        """
        raise NotImplementedError


class DataSource(AbstractDataSource):
    """ A trival but concrete implementation of AbstractDataSource which 
    handles basic functionality for manipulation of incoming timestamp 
    data.

    """
    #: The signal object used to manage subscriptions
    _data_changed = Signal()

    def __init__(self, publisher, buffer_size=1000000, sampler=None):
        """ Initialize a TimeStampDataSource.

        Parameters
        ----------
        publisher : AbstractPublisher
            A publisher instance which is publishing the feed we're
            interested in, at the rate at which we need it.

        buffer_size : int, optional
            The size of the data buffer to maintain. A full buffer is
            implemented as a sliding window across the incoming data.
            The default buffer size is 1e6.
        
        sampler : AbstractDataSampler, optional
            If provided, an instance of AbstractDataSampler which has
            the opportunity to subsample the data before it is 
            republished to subcribers.

        """
        self.buffer_size = buffer_size
        self.sampler = sampler
        self.publisher = publisher
        self._buffer = None

    def dispatch(self, data):
        """ The publisher callback which aggregates the new data and 
        emits the _data_changed signal.

        """
        buf = self._buffer
        max_size = self.buffer_size
        if len(data) >= max_size:
            buf = np.copy(data[-max_size:])
        else:
            if buf is None:
                buf = np.copy(data)
            else:
                offset = max_size - len(data)
                buf = np.hstack((buf[-offset:], data))
        sampler = self.sampler
        if sampler is not None:
            buf = sampler(buf)

        self._buffer = buf
        self._data_changed(buf)

    def bind(self):
        """ Bind the data source to the given publisher. This may only
        happen once.

        """
        self.publisher.subscribe(self.dispatch)

    def subscribe(self, callback):
        """ Subscribe the given callback to be notified when new
        data is available.

        """
        self._data_changed.connect(callback)

    def unsubscribe(self, callback):
        """ Unsubscribe a previously registered callback.

        """
        self._data_changed.disconnect(callback)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .wx_container import WXContainer

from ...components.tab import AbstractTkTab


class WXTab(WXContainer, AbstractTkTab):
    """ A wx implementation of the Tab component.
    
    """
    #--------------------------------------------------------------------------
    # Setup Methods 
    #--------------------------------------------------------------------------
    def initialize(self):
        """ Initialize the attributes of the tab.

        """
        super(WXTab, self).initialize()
        shell = self.shell_obj
        # We don't want to set the title during initialization on wx
        # because it will lead to segfaults (!). It's not a huge deal
        # however, because the title will be set when the tab is added
        # to the notebook.
        # self.set_title(shell.title)
        self.set_icon(shell.icon)

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_title_changed(self, title):
        """ The change handler for the 'title' attribute on the shell 
        object.

        """
        self.set_title(title)
    
    def shell_icon_changed(self, icon):
        """ The change handler for the 'icon' attribute on the shell
        object.

        """
        self.set_icon(icon)
    
    #--------------------------------------------------------------------------
    # Widget Update Methods 
    #--------------------------------------------------------------------------
    def set_title(self, title):
        """ Sets the title of this tab in the parent tab widget.

        """
        widget = self.widget
        tab_widget = self.shell_obj.parent.toolkit_widget
        for idx in range(tab_widget.GetPageCount()):
            if tab_widget.GetPage(idx) is widget:
                tab_widget.SetPageText(idx, title)
                break
     
    def set_icon(self, icon):
        """ Sets the icon of this tab in the parent tab widget.

        """
        # XXX handle icons
        pass




#------------------------------------------------------------------------------
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import sys

from ..qt.QtGui import QImage, QPixmap

from ....noncomponents.abstract_image import AbstractTkImage


class QtImage(AbstractTkImage):
    """ A Qt4 implementation of AbstractTkImage.
    
    """
    def __init__(self, qimage=None):
        """ Initialize a QtImage.

        Parameters
        ----------
        qimage : QImage instance or None, optional
            A QImage instance which holds the data. If None is passed, 
            then an empty QImage will be created.

        """
        if qimage is None:
            qimage = QImage()
        self._qimage = qimage
    
    #--------------------------------------------------------------------------
    # Abstract API Implementation
    #--------------------------------------------------------------------------
    @classmethod
    def from_data(cls, data, size):
        """ Construct an image from a bytearray of RGBA bytes.

        Parameters
        ----------
        data : bytearray
            A bytearray of the image data in RGBA32 format.

        size : (width, height)
            The width, height size tuple of the image.

        Returns
        -------
        results : AbstractTkImage
            An appropriate image instance.

        """
        # Shuffle the bytearray to match Qt's wonky ARGB format
        qt_array = bytearray(len(data))
        if sys.byteorder == 'little':
            r, g, b, a = (2, 1, 0, 3)
        else:
            r, g, b, a = (1, 2, 3, 0)
        qt_array[r::4] = data[0::4]
        qt_array[g::4] = data[1::4]
        qt_array[b::4] = data[2::4]
        qt_array[a::4] = data[3::4]

        w, h = size
        img_data = str(qt_array)
        qimg = QImage(img_data, w, h, QImage.Format_ARGB32)
        # Add a strong reference to the data so it doesn't get pulled out from
        # under the QImage object.
        qimg._data_keepalive_ptr = img_data
        return cls(qimg)

    @classmethod
    def from_file(cls, path, format=''):
        """ Read in the image data from a file.

        Parameters
        ----------
        path : string
            The path to the image file on disk.

        format : string or None, optional
            The image format of the data in the file. If not given,
            then the image format will be determined automatically
            if possible.
        
        Returns
        -------
        results : AbstractTkImage
            An appropriate image instance.

        """
        return cls(QImage(path, format))

    @property
    def size(self):
        """ The size of the image as a (width, height) tuple.
        
        Returns
        -------
        result : (width, height)
            The width, height size tuple of the image.

        """
        size = self._qimage.size()
        return (size.width(), size.height())
    
    def data(self):
        """ Returns the data for the image as a bytearray of RGBA bytes.
        
        Returns
        -------
        results : bytearray
            A bytearray of the image data in RGBA32 format.

        """
        qimg = self._qimage
        if qimg.format() != QImage.Format_ARGB32:
            qimg = qimg.convertToFormat(QImage.Format_ARGB32)

        # If bits() is None, then the image is invalid.
        buf = qimg.bits()
        if buf is None:
            return bytearray()

        # Qt's ARGB data format stores data as 0xAARRGGBB. On a little 
        # endian machine, the buffer of bytes will be ordered BBGGRRAA.
        out = bytearray(len(buf))
        if sys.byteorder == 'little':
            r, g, b, a = (2, 1, 0, 3)
        else:
            r, g, b, a = (1, 2, 3, 0)
        out[0::4] = buf[r::4]
        out[1::4] = buf[g::4]
        out[2::4] = buf[b::4]
        out[3::4] = buf[a::4]
        return out

    def scale(self, size):
        """ Returns a new version of this image scaled to the given size.
        
        Parameters
        ----------
        size : (width, height)
            The size of the scaled image.
        
        Returns
        -------
        results : AbstractTkImage
            A new image of the proper scaled size.
            
        """
        return QtImage(self._qimage.scaled(*size))

    #--------------------------------------------------------------------------
    # Toolkit Specific
    #--------------------------------------------------------------------------
    def as_QImage(self):
        """ Returns the internal QImage instance.

        """
        return self._qimage

    def as_QPixmap(self):
        """ Returns a QPixmap generated from the underlying QImage.
        
        This is provided as a convenience for qt backend components 
        which require a QPixmap rather than a QImage.
        
        """
        return QPixmap.fromImage(self._qimage)




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod

from .byteplay import (
    LOAD_NAME, CALL_FUNCTION, ROT_THREE, LOAD_CONST, LOAD_ATTR, ROT_TWO,
    BUILD_TUPLE, RETURN_VALUE,
)


#------------------------------------------------------------------------------
# Abstract Inverter
#------------------------------------------------------------------------------
class AbstractInverter(object):
    """ An abstract base class which defined the api for creating code
    inverter objects. An inverter is responsible for converting a code
    list supplied by an expression into the inverse of that expression
    which applies the new value onto the expression. For example:

    The expression:
                    getattr(foo, 'bar')
    
    Becomes:
                    setattr(foo, 'bar', value)
    
    Likewise, the expression:
                    [spam.ham[0] for spam in foo][0].bar

    Would be translated into:
                    [spam.ham[0] for spam in foo][0].bar = value

    The value, among other information, is provided by the expression 
    by inserting the following names into the scope, which can be 
    retrieved with the LOAD_NAME op-code:
        
        _[expr] : The expression object
        _[obj]  : The component object which triggered the change
        _[name] : The attribute on the component which changed
        _[old]  : The old value of the attribute.
        _[new]  : The new value of the attribute which should be
                  applied to the inverted expression.
    
    The generated expression should return True if it successfully 
    completed the operation, False otherwise.

    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_inverted_code(self, code_list):
        """ Generates the byteplay code operations which represent
        the inverted expression. If the inverter is unable to invert
        the given expression, it should return None.

        Parameters
        ----------
        code_list : list of (op_code, op_arg)
            The list of byteplay code operations for the expression.

        Returns
        -------
        result : list of (op_code, op_arg) or None
            A new list of code ops for the inverted expression, or 
            None if the inverter is unable to invert the expression.
            The generated code, when executed, should return True if
            the operation was successful, False otherwise. Exceptions
            raised by the user code should be propagated.

        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Abstract Attribute Inverter 
#------------------------------------------------------------------------------
class AbstractAttributeInverter(AbstractInverter):
    """ An AbstractInverter subclass which converts a bytecode expression
    ending in LOAD_ATTR into an appropriate expression which calls a 
    delegate to do the setattr operation.

    """
    def get_inverted_code(self, code_list):
        """ Generates and returns a new code list for expressions ending
        in a LOAD_ATTR op_code.

        """
        attr_code, attr_arg = code_list[-2]
        if attr_code == LOAD_ATTR:
            handler = self.get_setattr_handler()
            new_code = code_list[:-2]
            new_code.extend([
                (LOAD_CONST, handler),
                (ROT_TWO, None),
                (LOAD_CONST, attr_arg),
                (LOAD_NAME, '_[new]'),
                (CALL_FUNCTION, 0x0003),
                (RETURN_VALUE, None),
            ])
            return new_code

    @abstractmethod
    def get_setattr_handler(self):
        """ Returns a function which accepts three arguments: an object,
        an attribute name, and a value, and performs a setattr operation
        with the items. The function that performs the operation should 
        return True on success and False on failure. This method must be 
        implemented by subclasses.

        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Generic Attribute Inverter
#------------------------------------------------------------------------------
class GenericAttributeInverter(AbstractAttributeInverter):
    """ A concrete implementation of AbstractAttributeInverter which 
    performs the attribute operation using the builtin setattr.

    """
    @staticmethod
    def _attribute_inverter(obj, name, value):
        """ Performs a setattr on the object and returns True.

        """
        setattr(obj, name, value)
        return True

    def get_setattr_handler(self):
        """ Returns the builtin setattr function which is used to 
        perform the set attribute operation.

        """
        return self._attribute_inverter


#------------------------------------------------------------------------------
# Abstract Call Inverter
#------------------------------------------------------------------------------
class AbstractCallInverter(AbstractInverter):
    """ An AbstractInverter subclass which converts a bytecode expression
    ending in a CALL_FUNCTION into an appropriate expression which calls a 
    delegate to do the setattr operation.

    """
    @staticmethod
    def _call_wrapper(handler):
        """ A call inverter helper function which takes the handler
        and returns a closure which will unpack the stack arguments, 
        calls the handler, the return its value.

        """
        def closure(func_obj, arg_tuple, arg_spec, value):
            nargs = arg_spec & 0xFF
            args = arg_tuple[:nargs]
            kwargs = dict(zip(arg_tuple[nargs::2], arg_tuple[nargs+1::2]))
            return handler(func_obj, args, kwargs, value)
        return closure

    def get_inverted_code(self, code_list):
        """ Generates and returns a new code list for expressions ending
        in a CALL_FUNCTION op_code.

        """
        func_code, func_arg = code_list[-2]
        if func_code == CALL_FUNCTION:
            handler = self._call_wrapper(self.get_call_handler())
            new_code = code_list[:-2]
            n_stack_args = (func_arg & 0xFF) + 2 * ((func_arg >> 8) & 0xFF)
            new_code.extend([
                (BUILD_TUPLE, n_stack_args),
                (LOAD_CONST, handler),
                (ROT_THREE, None),
                (LOAD_CONST, func_arg),
                (LOAD_NAME, '_[new]'),
                (CALL_FUNCTION, 0x0004),
                (RETURN_VALUE, None),
            ])
            return new_code

    @abstractmethod
    def get_call_handler(self):
        """ Returns a function which accepts four arguments: the object
        that would have been called, the arguments being passed, the
        keyword arguments being passed, and the value that should be
        applied. The function that performs the operation should return
        True on success and False on failure. This method must be 
        implemented by subclasses.

        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Getattr Inverter
#------------------------------------------------------------------------------
class GetattrInverter(AbstractCallInverter):
    """ A concrete implementation of AbstractCallInverter which inverts
    a call to getattr into a call to setattr.

    """
    @staticmethod
    def _getattr_inverter(func_obj, args, kwargs, value):
        """ Performs the getattr inversion provided that the function
        being called is the builtin getattr with an appropriate arg
        spec.

        """
        if func_obj is getattr:
            if len(args) >= 2 and len(kwargs) == 0:
                setattr(args[0], args[1], value)
                return True
        return False

    def get_call_handler(self):
        """ Returns a function which accepts four arguments: the object
        that would have been called, the arguments being passed, the
        keyword arguments being passed, and the value that should be
        applied.

        """
        return self._getattr_inverter


#------------------------------------------------------------------------------
# Abstract Name Inverter
#------------------------------------------------------------------------------
class AbstractNameInverter(AbstractInverter):
    """ An AbstractInverter subclass which converts a bytecode expression
    which consists soley of a LOAD_NAME op_code into an appropriate 
    expression which calls a delegate to do the setattr operation.

    """
    def get_inverted_code(self, code_list):
        """ Generates and returns a new code list for expressions which
        consist solely of a LOAD_NAME op_code.

        """
        name_code, name_arg = code_list[-2]
        if name_code == LOAD_NAME and len(code_list) == 3:
            handler = self.get_name_handler()
            new_code = code_list[:-2]
            new_code.extend([
                (LOAD_CONST, handler),
                (LOAD_NAME, '_[expr]'),
                (LOAD_NAME, '_[obj]'),
                (LOAD_NAME, '_[name]'),
                (LOAD_CONST, name_arg),
                (LOAD_NAME, '_[new]'),
                (CALL_FUNCTION, 0x0005),
                (RETURN_VALUE, None),
            ])
            return new_code

    @abstractmethod
    def get_name_handler(self):
        """ Returns a function which accepts five arguments: the 
        expression object which owns this inverter, the object which 
        owns the attribute that changed, the name of the attribute that
        changed, the name on which to store the value, and the value 
        which should be stored. The function that performs the operation
        should return True on success and False on failure. This method 
        must be implemented by subclasses.

        """
        raise NotImplementedError


#------------------------------------------------------------------------------
# Generic Name Inverter
#------------------------------------------------------------------------------
class ImplicitAttrInverter(AbstractNameInverter):
    """ A concrete implementation of AbstractNameInverter which inverts
    a name access into a setattr operation on an implicit attribute
    in the component hierarchy.

    """
    @staticmethod
    def _name_inverter(expr, obj, attr, store_name, value):
        """ Performs the name inversion by walking the component tree
        looking for an object which has an attribute of the right
        name. Will return False if no valid target is found, if the 
        name references an identifier, or if the assignment would 
        result in a direct cycle.

        """
        if store_name in expr.identifiers or attr == store_name:
            return False
        while obj is not None:
            if hasattr(obj, store_name):
                setattr(obj, store_name, value)
                return True
            else:
                obj = obj.parent
        return False

    def get_name_handler(self):
        """ Returns a function which accepts five arguments: the 
        expression object which owns this inverter, the object which 
        owns the attribute that changed, the name of the attribute that
        changed, the new value of that attribute, and the name onto 
        which the value should be applied.

        """
        return self._name_inverter




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from .qt import QtGui
from .qt_abstract_item_view import QtAbstractItemView

from ...components.tree_view import AbstractTkTreeView


class QtTreeView(QtAbstractItemView, AbstractTkTreeView):
    """ A Qt implementation of TreeView.

    """
    #--------------------------------------------------------------------------
    # Setup methods
    #--------------------------------------------------------------------------
    def create(self, parent):
        """ Create the underlying QTableView control.

        """
        self.widget = QtGui.QTreeView(parent)

    def initialize(self):
        """ Initialize the widget with the attributes of this instance.

        """
        super(QtTreeView, self).initialize()
        self.set_header_visible(self.shell_obj.header_visible)

    #--------------------------------------------------------------------------
    # Implementation
    #--------------------------------------------------------------------------
    def shell_header_visible_changed(self, visible):
        """ The change handler for the 'header_visible' attribute of
        the shell object.

        """
        self.set_header_visible(visible)
    
    #--------------------------------------------------------------------------
    # Widget Update Methods
    #--------------------------------------------------------------------------
    def set_header_visible(self, visible):
        """ Sets the visibility of the header for the widget.

        """
        if visible:
            self.widget.setHeaderHidden(False)
        else:
            self.widget.setHeaderHidden(True)




# byteplay - Python bytecode assembler/disassembler.
# Copyright (C) 2006-2010 Noam Yorav-Raphael
# Homepage: http://code.google.com/p/byteplay
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

# Many thanks to Greg X for adding support for Python 2.6 and 2.7!

__version__ = '0.2'

__all__ = ['opmap', 'opname', 'opcodes',
           'cmp_op', 'hasarg', 'hasname', 'hasjrel', 'hasjabs',
           'hasjump', 'haslocal', 'hascompare', 'hasfree', 'hascode',
           'hasflow', 'getse',
           'Opcode', 'SetLineno', 'Label', 'isopcode', 'Code',
           'CodeList', 'printcodelist']

import opcode
from dis import findlabels
import types
from array import array
import operator
import itertools
import sys
import warnings
from cStringIO import StringIO

######################################################################
# Define opcodes and information about them

python_version = '.'.join(str(x) for x in sys.version_info[:2])
if python_version not in ('2.4', '2.5', '2.6', '2.7'):
    warnings.warn("byteplay doesn't support Python version "+python_version)

class Opcode(int):
    """An int which represents an opcode - has a nicer repr."""
    def __repr__(self):
        return opname[self]
    __str__ = __repr__

class CodeList(list):
    """A list for storing opcode tuples - has a nicer __str__."""
    def __str__(self):
        f = StringIO()
        printcodelist(self, f)
        return f.getvalue()

opmap = dict((name.replace('+', '_'), Opcode(code))
             for name, code in opcode.opmap.iteritems()
             if name != 'EXTENDED_ARG')
opname = dict((code, name) for name, code in opmap.iteritems())
opcodes = set(opname)

def globalize_opcodes():
    for name, code in opmap.iteritems():
        globals()[name] = code
        __all__.append(name)
globalize_opcodes()

cmp_op = opcode.cmp_op

hasarg = set(x for x in opcodes if x >= opcode.HAVE_ARGUMENT)
hasconst = set(Opcode(x) for x in opcode.hasconst)
hasname = set(Opcode(x) for x in opcode.hasname)
hasjrel = set(Opcode(x) for x in opcode.hasjrel)
hasjabs = set(Opcode(x) for x in opcode.hasjabs)
hasjump = hasjrel.union(hasjabs)
haslocal = set(Opcode(x) for x in opcode.haslocal)
hascompare = set(Opcode(x) for x in opcode.hascompare)
hasfree = set(Opcode(x) for x in opcode.hasfree)
hascode = set([MAKE_FUNCTION, MAKE_CLOSURE])

class _se:
    """Quick way of defining static stack effects of opcodes"""
    # Taken from assembler.py by Phillip J. Eby
    NOP       = 0,0

    POP_TOP   = 1,0
    ROT_TWO   = 2,2
    ROT_THREE = 3,3
    ROT_FOUR  = 4,4
    DUP_TOP   = 1,2

    UNARY_POSITIVE = UNARY_NEGATIVE = UNARY_NOT = UNARY_CONVERT = \
        UNARY_INVERT = GET_ITER = LOAD_ATTR = 1,1

    IMPORT_FROM = 1,2

    BINARY_POWER = BINARY_MULTIPLY = BINARY_DIVIDE = BINARY_FLOOR_DIVIDE = \
        BINARY_TRUE_DIVIDE = BINARY_MODULO = BINARY_ADD = BINARY_SUBTRACT = \
        BINARY_SUBSCR = BINARY_LSHIFT = BINARY_RSHIFT = BINARY_AND = \
        BINARY_XOR = BINARY_OR = COMPARE_OP = 2,1

    INPLACE_POWER = INPLACE_MULTIPLY = INPLACE_DIVIDE = \
        INPLACE_FLOOR_DIVIDE = INPLACE_TRUE_DIVIDE = INPLACE_MODULO = \
        INPLACE_ADD = INPLACE_SUBTRACT = INPLACE_LSHIFT = INPLACE_RSHIFT = \
        INPLACE_AND = INPLACE_XOR = INPLACE_OR = 2,1

    SLICE_0, SLICE_1, SLICE_2, SLICE_3 = \
        (1,1),(2,1),(2,1),(3,1)
    STORE_SLICE_0, STORE_SLICE_1, STORE_SLICE_2, STORE_SLICE_3 = \
        (2,0),(3,0),(3,0),(4,0)
    DELETE_SLICE_0, DELETE_SLICE_1, DELETE_SLICE_2, DELETE_SLICE_3 = \
        (1,0),(2,0),(2,0),(3,0)

    STORE_SUBSCR = 3,0
    DELETE_SUBSCR = STORE_ATTR = 2,0
    DELETE_ATTR = STORE_DEREF = 1,0
    PRINT_NEWLINE = 0,0
    PRINT_EXPR = PRINT_ITEM = PRINT_NEWLINE_TO = IMPORT_STAR = 1,0
    STORE_NAME = STORE_GLOBAL = STORE_FAST = 1,0
    PRINT_ITEM_TO = 2,0

    LOAD_LOCALS = LOAD_CONST = LOAD_NAME = LOAD_GLOBAL = LOAD_FAST = \
        LOAD_CLOSURE = LOAD_DEREF = BUILD_MAP = 0,1

    DELETE_FAST = DELETE_GLOBAL = DELETE_NAME = 0,0

    EXEC_STMT = 3,0
    BUILD_CLASS = 3,1

    STORE_MAP = MAP_ADD = 2,0
    SET_ADD = 1,0

    if   python_version == '2.4':
      YIELD_VALUE = 1,0
      IMPORT_NAME = 1,1
      LIST_APPEND = 2,0
    elif python_version == '2.5':
      YIELD_VALUE = 1,1
      IMPORT_NAME = 2,1
      LIST_APPEND = 2,0
    elif python_version == '2.6':
      YIELD_VALUE = 1,1
      IMPORT_NAME = 2,1
      LIST_APPEND = 2,0
    elif python_version == '2.7':
      YIELD_VALUE = 1,1
      IMPORT_NAME = 2,1
      LIST_APPEND = 1,0


_se = dict((op, getattr(_se, opname[op]))
           for op in opcodes
           if hasattr(_se, opname[op]))

hasflow = opcodes - set(_se) - \
          set([CALL_FUNCTION, CALL_FUNCTION_VAR, CALL_FUNCTION_KW,
               CALL_FUNCTION_VAR_KW, BUILD_TUPLE, BUILD_LIST,
               UNPACK_SEQUENCE, BUILD_SLICE, DUP_TOPX,
               RAISE_VARARGS, MAKE_FUNCTION, MAKE_CLOSURE])
if python_version == '2.7':
  hasflow = hasflow - set([BUILD_SET])

def getse(op, arg=None):
    """Get the stack effect of an opcode, as a (pop, push) tuple.

    If an arg is needed and is not given, a ValueError is raised.
    If op isn't a simple opcode, that is, the flow doesn't always continue
    to the next opcode, a ValueError is raised.
    """
    try:
        return _se[op]
    except KeyError:
        # Continue to opcodes with an effect that depends on arg
        pass

    if arg is None:
        raise ValueError, "Opcode stack behaviour depends on arg"

    def get_func_tup(arg, nextra):
        if arg > 0xFFFF:
            raise ValueError, "Can only split a two-byte argument"
        return (nextra + 1 + (arg & 0xFF) + 2*((arg >> 8) & 0xFF),
                1)

    if op == CALL_FUNCTION:
        return get_func_tup(arg, 0)
    elif op == CALL_FUNCTION_VAR:
        return get_func_tup(arg, 1)
    elif op == CALL_FUNCTION_KW:
        return get_func_tup(arg, 1)
    elif op == CALL_FUNCTION_VAR_KW:
        return get_func_tup(arg, 2)

    elif op == BUILD_TUPLE:
        return arg, 1
    elif op == BUILD_LIST:
        return arg, 1
    elif python_version == '2.7' and op == BUILD_SET:
        return arg, 1
    elif op == UNPACK_SEQUENCE:
        return 1, arg
    elif op == BUILD_SLICE:
        return arg, 1
    elif op == DUP_TOPX:
        return arg, arg*2
    elif op == RAISE_VARARGS:
        return 1+arg, 1
    elif op == MAKE_FUNCTION:
        return 1+arg, 1
    elif op == MAKE_CLOSURE:
        if python_version == '2.4':
            raise ValueError, "The stack effect of MAKE_CLOSURE depends on TOS"
        else:
            return 2+arg, 1
    else:
        raise ValueError, "The opcode %r isn't recognized or has a special "\
              "flow control" % op

class SetLinenoType(object):
    def __repr__(self):
        return 'SetLineno'
SetLineno = SetLinenoType()

class Label(object):
    pass

def isopcode(obj):
    """Return whether obj is an opcode - not SetLineno or Label"""
    return obj is not SetLineno and not isinstance(obj, Label)

# Flags from code.h
CO_OPTIMIZED              = 0x0001      # use LOAD/STORE_FAST instead of _NAME
CO_NEWLOCALS              = 0x0002      # only cleared for module/exec code
CO_VARARGS                = 0x0004
CO_VARKEYWORDS            = 0x0008
CO_NESTED                 = 0x0010      # ???
CO_GENERATOR              = 0x0020
CO_NOFREE                 = 0x0040      # set if no free or cell vars
CO_GENERATOR_ALLOWED      = 0x1000      # unused
# The future flags are only used on code generation, so we can ignore them.
# (It does cause some warnings, though.)
CO_FUTURE_DIVISION        = 0x2000
CO_FUTURE_ABSOLUTE_IMPORT = 0x4000
CO_FUTURE_WITH_STATEMENT  = 0x8000


######################################################################
# Define the Code class

class Code(object):
    """An object which holds all the information which a Python code object
    holds, but in an easy-to-play-with representation.

    The attributes are:

    Affecting action
    ----------------
    code - list of 2-tuples: the code
    freevars - list of strings: the free vars of the code (those are names
               of variables created in outer functions and used in the function)
    args - list of strings: the arguments of the code
    varargs - boolean: Does args end with a '*args' argument
    varkwargs - boolean: Does args end with a '**kwargs' argument
    newlocals - boolean: Should a new local namespace be created.
                (True in functions, False for module and exec code)

    Not affecting action
    --------------------
    name - string: the name of the code (co_name)
    filename - string: the file name of the code (co_filename)
    firstlineno - int: the first line number (co_firstlineno)
    docstring - string or None: the docstring (the first item of co_consts,
                if it's str or unicode)

    code is a list of 2-tuples. The first item is an opcode, or SetLineno, or a
    Label instance. The second item is the argument, if applicable, or None.
    code can be a CodeList instance, which will produce nicer output when
    being printed.
    """
    def __init__(self, code, freevars, args, varargs, varkwargs, newlocals,
                 name, filename, firstlineno, docstring):
        self.code = code
        self.freevars = freevars
        self.args = args
        self.varargs = varargs
        self.varkwargs = varkwargs
        self.newlocals = newlocals
        self.name = name
        self.filename = filename
        self.firstlineno = firstlineno
        self.docstring = docstring

    @staticmethod
    def _findlinestarts(code):
        """Find the offsets in a byte code which are start of lines in the
        source.

        Generate pairs (offset, lineno) as described in Python/compile.c.

        This is a modified version of dis.findlinestarts, which allows multiple
        "line starts" with the same line number.
        """
        byte_increments = [ord(c) for c in code.co_lnotab[0::2]]
        line_increments = [ord(c) for c in code.co_lnotab[1::2]]

        lineno = code.co_firstlineno
        addr = 0
        for byte_incr, line_incr in zip(byte_increments, line_increments):
            if byte_incr:
                yield (addr, lineno)
                addr += byte_incr
            lineno += line_incr
        yield (addr, lineno)

    @classmethod
    def from_code(cls, co):
        """Disassemble a Python code object into a Code object."""
        co_code = co.co_code
        labels = dict((addr, Label()) for addr in findlabels(co_code))
        linestarts = dict(cls._findlinestarts(co))
        cellfree = co.co_cellvars + co.co_freevars

        code = CodeList()
        n = len(co_code)
        i = 0
        extended_arg = 0
        while i < n:
            op = Opcode(ord(co_code[i]))
            if i in labels:
                code.append((labels[i], None))
            if i in linestarts:
                code.append((SetLineno, linestarts[i]))
            i += 1
            if op in hascode:
                lastop, lastarg = code[-1]
                if lastop != LOAD_CONST:
                    raise ValueError, \
                          "%s should be preceded by LOAD_CONST code" % op
                code[-1] = (LOAD_CONST, Code.from_code(lastarg))
            if op not in hasarg:
                code.append((op, None))
            else:
                arg = ord(co_code[i]) + ord(co_code[i+1])*256 + extended_arg
                extended_arg = 0
                i += 2
                if op == opcode.EXTENDED_ARG:
                    extended_arg = arg << 16
                elif op in hasconst:
                    code.append((op, co.co_consts[arg]))
                elif op in hasname:
                    code.append((op, co.co_names[arg]))
                elif op in hasjabs:
                    code.append((op, labels[arg]))
                elif op in hasjrel:
                    code.append((op, labels[i + arg]))
                elif op in haslocal:
                    code.append((op, co.co_varnames[arg]))
                elif op in hascompare:
                    code.append((op, cmp_op[arg]))
                elif op in hasfree:
                    code.append((op, cellfree[arg]))
                else:
                    code.append((op, arg))

        varargs = bool(co.co_flags & CO_VARARGS)
        varkwargs = bool(co.co_flags & CO_VARKEYWORDS)
        newlocals = bool(co.co_flags & CO_NEWLOCALS)
        args = co.co_varnames[:co.co_argcount + varargs + varkwargs]
        if co.co_consts and isinstance(co.co_consts[0], basestring):
            docstring = co.co_consts[0]
        else:
            docstring = None
        return cls(code = code,
                   freevars = co.co_freevars,
                   args = args,
                   varargs = varargs,
                   varkwargs = varkwargs,
                   newlocals = newlocals,
                   name = co.co_name,
                   filename = co.co_filename,
                   firstlineno = co.co_firstlineno,
                   docstring = docstring,
                   )

    def __eq__(self, other):
        if (self.freevars != other.freevars or
            self.args != other.args or
            self.varargs != other.varargs or
            self.varkwargs != other.varkwargs or
            self.newlocals != other.newlocals or
            self.name != other.name or
            self.filename != other.filename or
            self.firstlineno != other.firstlineno or
            self.docstring != other.docstring or
            len(self.code) != len(other.code)
            ):
            return False

        # Compare code. This isn't trivial because labels should be matching,
        # not equal.
        labelmapping = {}
        for (op1, arg1), (op2, arg2) in itertools.izip(self.code, other.code):
            if isinstance(op1, Label):
                if labelmapping.setdefault(op1, op2) is not op2:
                    return False
            else:
                if op1 != op2:
                    return False
                if op1 in hasjump:
                    if labelmapping.setdefault(arg1, arg2) is not arg2:
                        return False
                elif op1 in hasarg:
                    if arg1 != arg2:
                        return False
        return True

    def _compute_flags(self):
        opcodes = set(op for op, arg in self.code if isopcode(op))

        optimized = (STORE_NAME not in opcodes and
                     LOAD_NAME not in opcodes and
                     DELETE_NAME not in opcodes)
        generator = (YIELD_VALUE in opcodes)
        nofree = not (opcodes.intersection(hasfree))

        flags = 0
        if optimized: flags |= CO_OPTIMIZED
        if self.newlocals: flags |= CO_NEWLOCALS
        if self.varargs: flags |= CO_VARARGS
        if self.varkwargs: flags |= CO_VARKEYWORDS
        if generator: flags |= CO_GENERATOR
        if nofree: flags |= CO_NOFREE
        return flags

    def _compute_stacksize(self):
        """Get a code list, compute its maximal stack usage."""
        # This is done by scanning the code, and computing for each opcode
        # the stack state at the opcode.
        code = self.code

        # A mapping from labels to their positions in the code list
        label_pos = dict((op, pos)
                         for pos, (op, arg) in enumerate(code)
                         if isinstance(op, Label))

        # sf_targets are the targets of SETUP_FINALLY opcodes. They are recorded
        # because they have special stack behaviour. If an exception was raised
        # in the block pushed by a SETUP_FINALLY opcode, the block is popped
        # and 3 objects are pushed. On return or continue, the block is popped
        # and 2 objects are pushed. If nothing happened, the block is popped by
        # a POP_BLOCK opcode and 1 object is pushed by a (LOAD_CONST, None)
        # operation.
        #
        # Our solution is to record the stack state of SETUP_FINALLY targets
        # as having 3 objects pushed, which is the maximum. However, to make
        # stack recording consistent, the get_next_stacks function will always
        # yield the stack state of the target as if 1 object was pushed, but
        # this will be corrected in the actual stack recording.

        sf_targets = set(label_pos[arg]
                         for op, arg in code
                         if op == SETUP_FINALLY)

        # What we compute - for each opcode, its stack state, as an n-tuple.
        # n is the number of blocks pushed. For each block, we record the number
        # of objects pushed.
        stacks = [None] * len(code)

        def get_next_stacks(pos, curstack):
            """Get a code position and the stack state before the operation
            was done, and yield pairs (pos, curstack) for the next positions
            to be explored - those are the positions to which you can get
            from the given (pos, curstack).

            If the given position was already explored, nothing will be yielded.
            """
            op, arg = code[pos]

            if isinstance(op, Label):
                # We should check if we already reached a node only if it is
                # a label.
                if pos in sf_targets:
                    curstack = curstack[:-1] + (curstack[-1] + 2,)
                if stacks[pos] is None:
                    stacks[pos] = curstack
                else:
                    if stacks[pos] != curstack:
                        raise ValueError, "Inconsistent code"
                    return

            def newstack(n):
                # Return a new stack, modified by adding n elements to the last
                # block
                if curstack[-1] + n < 0:
                    raise ValueError, "Popped a non-existing element"
                return curstack[:-1] + (curstack[-1]+n,)

            if not isopcode(op):
                # label or SetLineno - just continue to next line
                yield pos+1, curstack

            elif op in (STOP_CODE, RETURN_VALUE, RAISE_VARARGS):
                # No place in particular to continue to
                pass

            elif op == MAKE_CLOSURE and python_version == '2.4':
                # This is only relevant in Python 2.4 - in Python 2.5 the stack
                # effect of MAKE_CLOSURE can be calculated from the arg.
                # In Python 2.4, it depends on the number of freevars of TOS,
                # which should be a code object.
                if pos == 0:
                    raise ValueError, \
                          "MAKE_CLOSURE can't be the first opcode"
                lastop, lastarg = code[pos-1]
                if lastop != LOAD_CONST:
                    raise ValueError, \
                          "MAKE_CLOSURE should come after a LOAD_CONST op"
                try:
                    nextrapops = len(lastarg.freevars)
                except AttributeError:
                    try:
                        nextrapops = len(lastarg.co_freevars)
                    except AttributeError:
                        raise ValueError, \
                              "MAKE_CLOSURE preceding const should "\
                              "be a code or a Code object"

                yield pos+1, newstack(-arg-nextrapops)

            elif op not in hasflow:
                # Simple change of stack
                pop, push = getse(op, arg)
                yield pos+1, newstack(push - pop)

            elif op in (JUMP_FORWARD, JUMP_ABSOLUTE):
                # One possibility for a jump
                yield label_pos[arg], curstack

            elif python_version < '2.7' and op in (JUMP_IF_FALSE, JUMP_IF_TRUE):
                # Two possibilities for a jump
                yield label_pos[arg], curstack
                yield pos+1, curstack

            elif python_version >= '2.7' and op in (POP_JUMP_IF_FALSE, POP_JUMP_IF_TRUE):
                # Two possibilities for a jump
                yield label_pos[arg], newstack(-1)
                yield pos+1, newstack(-1)

            elif python_version >= '2.7' and op in (JUMP_IF_TRUE_OR_POP, JUMP_IF_FALSE_OR_POP):
                # Two possibilities for a jump
                yield label_pos[arg], curstack
                yield pos+1, newstack(-1)

            elif op == FOR_ITER:
                # FOR_ITER pushes next(TOS) on success, and pops TOS and jumps
                # on failure
                yield label_pos[arg], newstack(-1)
                yield pos+1, newstack(1)

            elif op == BREAK_LOOP:
                # BREAK_LOOP jumps to a place specified on block creation, so
                # it is ignored here
                pass

            elif op == CONTINUE_LOOP:
                # CONTINUE_LOOP jumps to the beginning of a loop which should
                # already ave been discovered, but we verify anyway.
                # It pops a block.
                if python_version == '2.6':
                  pos, stack = label_pos[arg], curstack[:-1]
                  if stacks[pos] != stack: #this could be a loop with a 'with' inside
                    yield pos, stack[:-1] + (stack[-1]-1,)
                  else:
                    yield pos, stack
                else:
                  yield label_pos[arg], curstack[:-1]

            elif op == SETUP_LOOP:
                # We continue with a new block.
                # On break, we jump to the label and return to current stack
                # state.
                yield label_pos[arg], curstack
                yield pos+1, curstack + (0,)

            elif op == SETUP_EXCEPT:
                # We continue with a new block.
                # On exception, we jump to the label with 3 extra objects on
                # stack
                yield label_pos[arg], newstack(3)
                yield pos+1, curstack + (0,)

            elif op == SETUP_FINALLY:
                # We continue with a new block.
                # On exception, we jump to the label with 3 extra objects on
                # stack, but to keep stack recording consistent, we behave as
                # if we add only 1 object. Extra 2 will be added to the actual
                # recording.
                yield label_pos[arg], newstack(1)
                yield pos+1, curstack + (0,)

            elif python_version == '2.7' and op == SETUP_WITH:
                yield label_pos[arg], curstack
                yield pos+1, newstack(-1) + (1,)

            elif op == POP_BLOCK:
                # Just pop the block
                yield pos+1, curstack[:-1]

            elif op == END_FINALLY:
                # Since stack recording of SETUP_FINALLY targets is of 3 pushed
                # objects (as when an exception is raised), we pop 3 objects.
                yield pos+1, newstack(-3)

            elif op == WITH_CLEANUP:
                # Since WITH_CLEANUP is always found after SETUP_FINALLY
                # targets, and the stack recording is that of a raised
                # exception, we can simply pop 1 object and let END_FINALLY
                # pop the remaining 3.
                if python_version == '2.7':
                  yield pos+1, newstack(2)
                else:
                  yield pos+1, newstack(-1)

            else:
                assert False, "Unhandled opcode: %r" % op


        # Now comes the calculation: open_positions holds positions which are
        # yet to be explored. In each step we take one open position, and
        # explore it by adding the positions to which you can get from it, to
        # open_positions. On the way, we update maxsize.
        # open_positions is a list of tuples: (pos, stack state)
        maxsize = 0
        open_positions = [(0, (0,))]
        while open_positions:
            pos, curstack = open_positions.pop()
            maxsize = max(maxsize, sum(curstack))
            open_positions.extend(get_next_stacks(pos, curstack))

        return maxsize

    def to_code(self):
        """Assemble a Python code object from a Code object."""
        co_argcount = len(self.args) - self.varargs - self.varkwargs
        co_stacksize = self._compute_stacksize()
        co_flags = self._compute_flags()

        co_consts = [self.docstring]
        co_names = []
        co_varnames = list(self.args)

        co_freevars = tuple(self.freevars)

        # We find all cellvars beforehand, for two reasons:
        # 1. We need the number of them to construct the numeric argument
        #    for ops in "hasfree".
        # 2. We need to put arguments which are cell vars in the beginning
        #    of co_cellvars
        cellvars = set(arg for op, arg in self.code
                       if isopcode(op) and op in hasfree
                       and arg not in co_freevars)
        co_cellvars = [x for x in self.args if x in cellvars]

        def index(seq, item, eq=operator.eq, can_append=True):
            """Find the index of item in a sequence and return it.
            If it is not found in the sequence, and can_append is True,
            it is appended to the sequence.

            eq is the equality operator to use.
            """
            for i, x in enumerate(seq):
                if eq(x, item):
                    return i
            else:
                if can_append:
                    seq.append(item)
                    return len(seq) - 1
                else:
                    raise IndexError, "Item not found"

        # List of tuples (pos, label) to be filled later
        jumps = []
        # A mapping from a label to its position
        label_pos = {}
        # Last SetLineno
        lastlineno = self.firstlineno
        lastlinepos = 0

        co_code = array('B')
        co_lnotab = array('B')
        for i, (op, arg) in enumerate(self.code):
            if isinstance(op, Label):
                label_pos[op] = len(co_code)

            elif op is SetLineno:
                incr_lineno = arg - lastlineno
                incr_pos = len(co_code) - lastlinepos
                lastlineno = arg
                lastlinepos = len(co_code)

                if incr_lineno == 0 and incr_pos == 0:
                    co_lnotab.append(0)
                    co_lnotab.append(0)
                else:
                    while incr_pos > 255:
                        co_lnotab.append(255)
                        co_lnotab.append(0)
                        incr_pos -= 255
                    while incr_lineno > 255:
                        co_lnotab.append(incr_pos)
                        co_lnotab.append(255)
                        incr_pos = 0
                        incr_lineno -= 255
                    if incr_pos or incr_lineno:
                        co_lnotab.append(incr_pos)
                        co_lnotab.append(incr_lineno)

            elif op == opcode.EXTENDED_ARG:
                raise ValueError, "EXTENDED_ARG not supported in Code objects"

            elif not op in hasarg:
                co_code.append(op)

            else:
                if op in hasconst:
                    if isinstance(arg, Code) and i < len(self.code)-1 and \
                       self.code[i+1][0] in hascode:
                        arg = arg.to_code()
                    arg = index(co_consts, arg, operator.is_)
                elif op in hasname:
                    arg = index(co_names, arg)
                elif op in hasjump:
                    # arg will be filled later
                    jumps.append((len(co_code), arg))
                    arg = 0
                elif op in haslocal:
                    arg = index(co_varnames, arg)
                elif op in hascompare:
                    arg = index(cmp_op, arg, can_append=False)
                elif op in hasfree:
                    try:
                        arg = index(co_freevars, arg, can_append=False) \
                              + len(cellvars)
                    except IndexError:
                        arg = index(co_cellvars, arg)
                else:
                    # arg is ok
                    pass

                if arg > 0xFFFF:
                    co_code.append(opcode.EXTENDED_ARG)
                    co_code.append((arg >> 16) & 0xFF)
                    co_code.append((arg >> 24) & 0xFF)
                co_code.append(op)
                co_code.append(arg & 0xFF)
                co_code.append((arg >> 8) & 0xFF)

        for pos, label in jumps:
            jump = label_pos[label]
            if co_code[pos] in hasjrel:
                jump -= pos+3
            if jump > 0xFFFF:
                raise NotImplementedError, "Extended jumps not implemented"
            co_code[pos+1] = jump & 0xFF
            co_code[pos+2] = (jump >> 8) & 0xFF

        co_code = co_code.tostring()
        co_lnotab = co_lnotab.tostring()

        co_consts = tuple(co_consts)
        co_names = tuple(co_names)
        co_varnames = tuple(co_varnames)
        co_nlocals = len(co_varnames)
        co_cellvars = tuple(co_cellvars)

        return types.CodeType(co_argcount, co_nlocals, co_stacksize, co_flags,
                              co_code, co_consts, co_names, co_varnames,
                              self.filename, self.name, self.firstlineno, co_lnotab,
                              co_freevars, co_cellvars)


def printcodelist(codelist, to=sys.stdout):
    """Get a code list. Print it nicely."""

    labeldict = {}
    pendinglabels = []
    for i, (op, arg) in enumerate(codelist):
        if isinstance(op, Label):
            pendinglabels.append(op)
        elif op is SetLineno:
            pass
        else:
            while pendinglabels:
                labeldict[pendinglabels.pop()] = i

    lineno = None
    islabel = False
    for i, (op, arg) in enumerate(codelist):
        if op is SetLineno:
            lineno = arg
            print >> to
            continue

        if isinstance(op, Label):
            islabel = True
            continue

        if lineno is None:
            linenostr = ''
        else:
            linenostr = str(lineno)
            lineno = None

        if islabel:
            islabelstr = '>>'
            islabel = False
        else:
            islabelstr = ''

        if op in hasconst:
            argstr = repr(arg)
        elif op in hasjump:
            try:
                argstr = 'to ' + str(labeldict[arg])
            except KeyError:
                argstr = repr(arg)
        elif op in hasarg:
            argstr = str(arg)
        else:
            argstr = ''

        print >> to, '%3s     %2s %4d %-20s %s' % (
            linenostr,
            islabelstr,
            i,
            op,
            argstr)

def recompile(filename):
    """Create a .pyc by disassembling the file and assembling it again, printing
    a message that the reassembled file was loaded."""
    # Most of the code here based on the compile.py module.
    import os
    import imp
    import marshal
    import struct

    f = open(filename, 'U')
    try:
        timestamp = long(os.fstat(f.fileno()).st_mtime)
    except AttributeError:
        timestamp = long(os.stat(filename).st_mtime)
    codestring = f.read()
    f.close()
    if codestring and codestring[-1] != '\n':
        codestring = codestring + '\n'
    try:
        codeobject = compile(codestring, filename, 'exec')
    except SyntaxError:
        print >> sys.stderr, "Skipping %s - syntax error." % filename
        return
    cod = Code.from_code(codeobject)
    message = "reassembled %r imported.\n" % filename
    cod.code[:0] = [ # __import__('sys').stderr.write(message)
        (LOAD_GLOBAL, '__import__'),
        (LOAD_CONST, 'sys'),
        (CALL_FUNCTION, 1),
        (LOAD_ATTR, 'stderr'),
        (LOAD_ATTR, 'write'),
        (LOAD_CONST, message),
        (CALL_FUNCTION, 1),
        (POP_TOP, None),
        ]
    codeobject2 = cod.to_code()
    fc = open(filename+'c', 'wb')
    fc.write('\0\0\0\0')
    fc.write(struct.pack('<l', timestamp))
    marshal.dump(codeobject2, fc)
    fc.flush()
    fc.seek(0, 0)
    fc.write(imp.get_magic())
    fc.close()

def recompile_all(path):
    """recursively recompile all .py files in the directory"""
    import os
    if os.path.isdir(path):
        for root, dirs, files in os.walk(path):
            for name in files:
                if name.endswith('.py'):
                    filename = os.path.abspath(os.path.join(root, name))
                    print >> sys.stderr, filename
                    recompile(filename)
    else:
        filename = os.path.abspath(path)
        recompile(filename)

def main():
    import os
    if len(sys.argv) != 2 or not os.path.exists(sys.argv[1]):
        print """\
Usage: %s dir

Search recursively for *.py in the given directory, disassemble and assemble
them, adding a note when each file is imported.

Use it to test byteplay like this:
> byteplay.py Lib
> make test

Some FutureWarnings may be raised, but that's expected.

Tip: before doing this, check to see which tests fail even without reassembling
them...
""" % sys.argv[0]
        sys.exit(1)
    recompile_all(sys.argv[1])

if __name__ == '__main__':
    main()



#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from unittest import expectedFailure

from enaml.backends.qt.qt import QtCore
from .qt_test_assistant import QtTestAssistant
from .. import group_box

# The alignment object we get back from widget.alignment() is not
# hashable. So we convert the flags to ints.
QT_2_ENAML_ALIGNMENTS = {int(QtCore.Qt.AlignLeft): 'left',
                         int(QtCore.Qt.AlignRight): 'right',
                         int(QtCore.Qt.AlignHCenter): 'center'}


class TestQtGroupBox(QtTestAssistant, group_box.TestGroupBox):

    # This test is an expected failure since calling process
    # events on qt still doesn't seem to empty the relayout
    # queue which has call-later events waiting. Without
    # those being processed, the title update never happens.
    @expectedFailure
    def test_title_changed(self):
        super(TestGroupBox, self).test_title_changed(self)

    def get_title(self, component, widget):
        """ Returns the title text from the tookit widget

        """
        # The title is set on a deferred call, so we need to pump
        # the event loop a bit to get the title to change.
        # XXX this doesn't work, see above comment.
        self.toolkit.app.process_events()
        return widget.title()

    # This test is an expected failure since calling process
    # events on qt still doesn't seem to empty the relayout
    # queue which has call-later events waiting. Without
    # those being processed, the flat update never happens.
    @expectedFailure
    def test_flat_style_changed(self):
        super(TestGroupBox, self).test_flat_style_changed(self)

    def get_flat(self, component, widget):
        """ Returns the flat style status from the tookit widget

        """
        # The flat is set on a deferred call, so we need to pump
        # the event loop a bit to get the title to change.
        # XXX this doesn't work, see above comment.
        self.toolkit.app.process_events()
        return widget.isFlat()

    def get_title_align(self, component, widget):
        """ Returns the title aligment in the tookit widget

        """
        alignment = int(widget.alignment())
        return QT_2_ENAML_ALIGNMENTS[alignment]


#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import (
    HasStrictTraits, List, Property, Instance, Bool, on_trait_change,
    Tuple, Either
)

from .box_model import BoxModel, PaddingBoxModel
from .geometry import Box

from ..core.trait_types import CoercingInstance
from ..enums import PolicyEnum


def _get_from_box_model(self, name):
    """ Property getter for all attributes that come from the box model.

    """
    return getattr(self._box_model, name)


class Constrainable(HasStrictTraits):
    """ A mixin class which adds the facilities necessary for a component
    to be managed by a constraints-based layout manager. This mixin is
    expecting to have BaseComponent in the mro.

    """
    #: The list of user-specified constraints or constraint-generating
    #: objects for this component.
    constraints = List
    
    #: Whether or not the component is visible. An invisible component
    #: will be ignored by constraints layout helpers.
    visible = Bool(True)

    #: A read-only symbolic object that represents the left boundary of 
    #: the component
    left = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the top boundary of 
    #: the component
    top = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the width of the 
    #: component
    width = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the height of the 
    #: component
    height = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the right boundary 
    #: of the component
    right = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the bottom boundary 
    #: of the component
    bottom = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the vertical center 
    #: of the component
    v_center = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the horizontal 
    #: center of the component
    h_center = Property(fget=_get_from_box_model)

    #: The private storage the box model instance for this component.
    _box_model = Instance(BoxModel)
    def __box_model_default(self):
        return BoxModel(self)

    #--------------------------------------------------------------------------
    # Change Handlers
    #--------------------------------------------------------------------------
    @on_trait_change('constraints, constraints_items')
    def _on_constrainable_deps_changed(self):
        """ A change handler which requests a relayout when the constraints
        change, provided that the component is initialized.

        """
        if self.initialized:
            self.request_relayout()
    
    #--------------------------------------------------------------------------
    # Geometry Methods
    #--------------------------------------------------------------------------
    def update_layout_geometry(self, dx, dy):
        """ An abstract method which must be implemented by subclasses.
        This method is called during a layout pass when the symbolic
        constraint variables of the component are filled with their
        solved values. The component should use the opportunity to
        update its layout geometry with the updated information.

        Parameters
        ----------
        dx : int
            The x-direction offset of the parent of this component from
            the root component on which the solved dimensions are based.
        
        dy : int
            The y-direction offset of the parent of this component from
            the root component on which the solved dimensions are based.
        
        Returns
        -------
        result : (x, y)
            The solved (x, y) position of this component relative to
            the root component on which the solved dimensions are bsaed.

        """
        raise NotImplementedError
    
    #--------------------------------------------------------------------------
    # Constraint Handling
    #--------------------------------------------------------------------------
    def hard_constraints(self):
        """ Returns the list of required symbolic constraints for the 
        component. These are constraints that apply to both the internal
        layout computations of a component as well as any calculations
        of containers which may parent this component. The default 
        implementation constrains the variables 'left', 'right', 
        'width', and 'height' to >= 0 required.

        """
        cns = [
            self.left >= 0, self.top >= 0, 
            self.width >= 0, self.height >= 0,
        ]
        return cns

    def size_hint_constraints(self):
        """ Returns the list of constraints which are related to the
        preferred size of the component. These constraints should be
        made independent of the 'hard_constraints' and are only of
        use in the calculations of potential parent containers. The 
        default implementation of this method returns an empty list.

        """
        return []
            
    def component_constraints(self):
        """ A set of constraints that should always be applied to this 
        type of component. The constraints generated by this method
        are use in the internal layout calculations of a component. 
        The default implementation returns an empty list.

        """
        return []

    def user_constraints(self):
        """ Returns the list of constraints specified by the user or the
        list of constraints computed by 'default_user_constraints' if the
        user has not supplied their own list. The list is used for 
        internal layout calculations.

        """
        cns = self.constraints
        if not cns:
            cns = self.default_user_constraints()
        return cns

    def default_user_constraints(self):
        """ Constraints to use if the constraints trait is an empty list.
        This default implementation returns an empty list.

        """
        return []


class PaddingConstraints(Constrainable):
    """ A Constrainable subclass which adds padding to the box model.

    """
    #: A read-only symbolic object that represents the internal left 
    #: padding of the container.
    padding_left = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal right 
    #: padding of the container.
    padding_right = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal top 
    #: padding of the container.
    padding_top = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal bottom 
    #: padding of the container.
    padding_bottom = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal left 
    #: boundary of the content area of the container.
    contents_left = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal right 
    #: boundary of the content area of the container.
    contents_right = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal top 
    #: boundary of the content area of the container.
    contents_top = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal bottom 
    #: boundary of the content area of the container.
    contents_bottom = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal width of
    #: the content area of the container.
    contents_width = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal height of
    #: the content area of the container.
    contents_height = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal center 
    #: along the vertical direction the content area of the container.
    contents_v_center = Property(fget=_get_from_box_model)

    #: A read-only symbolic object that represents the internal center 
    #: along the horizontal direction of the content area of the container.
    contents_h_center = Property(fget=_get_from_box_model)

    #: A box object which holds the padding for this component. The 
    #: default padding is Box(0, 0, 0, 0).
    padding = CoercingInstance(Box, default=Box(0, 0, 0, 0))

    #: The PolicyEnum for the strength with which to enforce the padding.
    #: This can be a single policy value to apply to everything, or a 
    #: 4-tuple of policies to apply to the individual padding.           
    padding_strength = Either(
        PolicyEnum, Tuple(PolicyEnum, PolicyEnum, PolicyEnum, PolicyEnum),
        default = 'required'
    )

    #: The private storage the box model instance for this component. 
    _box_model = Instance(PaddingBoxModel, ())
    def __box_model_default(self):
        return PaddingBoxModel(self)
    
    #--------------------------------------------------------------------------
    # Change Handlers
    #--------------------------------------------------------------------------
    @on_trait_change('padding', 'padding_strength')
    def _on_padding_deps_changed(self):
        """ A change handler which requests a relayout if the padding or
        padding strengths change, provided that the component is fully
        initialized.

        """
        if self.initialized:
            self.request_relayout()
    
    #--------------------------------------------------------------------------
    # Constraint Handling
    #--------------------------------------------------------------------------
    def padding_constraints(self):
        """ Returns the list of symbolic constraints for the padding of 
        the component. These constraints apply to the internal layout 
        calculations of an object. The default implementation constrains 
        the padding according the values in the 'padding' attribute and
        the strengths in 'padding_strength' attribute. It also places a 
        required constraint >= 0 on every padding element.

        """
        cns = []
        padding = self.padding
        tags = ('top', 'right', 'bottom', 'left')
        strengths = self.padding_strength
        if isinstance(strengths, basestring):
            strengths = (strengths,) * 4
        for tag, strength, padding in zip(tags, strengths, padding):
            tag = 'padding_' + tag
            sym = getattr(self, tag)
            cns.append(sym >= 0)
            if strength != 'ignore':
                cns.append((sym == padding) | strength)
        return cns




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
import functools


DEPRECATED_FUNCTION = ('The %s function is deprecated and will be removed in '
                       'future versions of Enaml.')


def deprecated(func):
    """ A decorator which will raise a DeprecationWarning when the 
    wrapped function is executed.

    """
    @functools.wraps(func)
    def closure(*args, **kwargs):
        import warnings
        warnings.warn(DEPRECATED_FUNCTION % func.__name__, DeprecationWarning)
        return func(*args, **kwargs)
    return closure




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import HasTraits, Str, Range, on_trait_change


class Person(HasTraits):
    """ A simple class representing a person object.

    """
    last_name = Str

    first_name = Str

    age = Range(low=0)

    @on_trait_change('age')
    def debug_print(self):
        """ Prints out a debug message whenever the person's age changes.

        """
        templ = "{first} {last} is {age} years old."
        print templ.format(first=self.first_name, 
                           last=self.last_name, 
                           age=self.age)


if __name__ == '__main__':
    import enaml
    with enaml.imports():
        from person_view import PersonView
    
    john = Person(first_name='John', last_name='Doe', age=42)
    
    view = PersonView(person=john)
    view.show()




#------------------------------------------------------------------------------
#  Copyright (c) 2011, Enthought, Inc.
#  All rights reserved.
#------------------------------------------------------------------------------
from traits.api import TraitError

from .enaml_test_case import EnamlTestCase, required_method


class TestGroupBox(EnamlTestCase):
    """ Logic for testing labels.

    Tooklit testcases need to provide the following methods:

    Abstract Methods
    ----------------
    get_title
        Returns the group title from the tookit widget

    get_flat
        Returns the flat border style from the tookit widget

    get_title_align
        Returns the title align style from the tookit widget

    Notes
    -----
    All the provided methods need to support the extented signature
    <method>(component, widget).

    """
    def setUp(self):
        """ Set up label tests.

        """
        enaml_source = """
enamldef MainView(MainWindow):
    name = 'win'
    Container:
        constraints = [horizontal(left, gb, right), vertical(top, gb, bottom),
                       vertical(label1, label2), align('left', label1, label2)]
        GroupBox:
            id: gb
            name = 'gb'
            title = 'MyGroup'
            flat = True
            title_align = 'center'
            Label:
                id: label1
                text = 'foofoofoofoofoofoofoofoofoofoofoofoo'
            Label:
                id: label2
                text = 'barbarbarbarbarbarbarbarbarbar'
"""
        self.view = self.parse_and_create(enaml_source)
        self.component = self.component_by_name(self.view, 'gb')
        self.widget = self.component.toolkit_widget

    def test_initialization(self):
        """ Test initialization

        Test that the atrributes of the GroupBox have the
        correct values and these values are insync with the information
        in the toolkit widget.

        """
        component = self.component
        self.assertEnamlInSync(component, 'title', 'MyGroup')
        self.assertEnamlInSync(component, 'flat', True)
        self.assertEnamlInSync(component, 'title_align', 'center')

    # This is an expected failure on Qt since the call later event
    # do not get process without the event loop running.
    def test_title_changed(self):
        """ Change the title text of the GroupBox.

        """
        component = self.component
        component.title = "New title"
        self.assertEnamlInSync(component, 'title', 'New title')

    # This is an expected failure on Qt since the call later event
    # do not get process without the event loop running.
    def test_flat_style_changed(self):
        """ Change the flat style border of the GroupBox.

        """
        component = self.component
        component.flat = False
        self.assertEnamlInSync(component, 'flat', False)

    def test_title_align_changed(self):
        """ Change the title alignment.

        """
        component = self.component
        component.title_align = 'right'
        self.assertEnamlInSync(component, 'title_align', 'right')
        component.title_align = 'left'
        self.assertEnamlInSync(component, 'title_align', 'left')
        component.title_align = 'center'
        self.assertEnamlInSync(component, 'title_align', 'center')

    def test_invalid_alignment(self):
        """ Change the title alignment to an invalid value.

        """
        component = self.component
        with self.assertRaises(TraitError):
            component.title_align = 'almost center'
        self.assertEnamlInSync(component, 'title_align', 'center')

    #--------------------------------------------------------------------------
    # Abstract methods
    #--------------------------------------------------------------------------
    @required_method
    def get_title(self, widget):
        """ Returns the title text from the tookit widget

        """
        pass

    @required_method
    def get_flat(self, widget):
        """ Returns the flat style status from the tookit widget

        """
        pass

    @required_method
    def get_title_align(self, widget):
        """ Returns the title aligment in the tookit widget

        """
        pass

